# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'logViewWindowTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.15.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(800, 600)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 800, 25))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionLoad = QtWidgets.QAction(MainWindow)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/open-folder.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionLoad.setIcon(icon)
        self.actionLoad.setObjectName("actionLoad")
        self.menuFile.addAction(self.actionLoad)
        self.menubar.addAction(self.menuFile.menuAction())

        self.retranslateUi(MainWindow)
        self.actionLoad.triggered.connect(MainWindow.load_file)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.menuFile.setTitle(_translate("MainWindow", "Open"))
        self.actionLoad.setText(_translate("MainWindow", "File"))
from kamzik3 import resource_kamzik3_rc



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'consoleTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.10.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(1009, 492)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.input_console_text = QtWidgets.QTextEdit(Form)
        self.input_console_text.setReadOnly(True)
        self.input_console_text.setObjectName("input_console_text")
        self.gridLayout.addWidget(self.input_console_text, 3, 0, 1, 1)
        self.label_console_header = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_console_header.sizePolicy().hasHeightForWidth())
        self.label_console_header.setSizePolicy(sizePolicy)
        self.label_console_header.setStyleSheet("font-weight:bold;background:#f6b442;padding:5")
        self.label_console_header.setObjectName("label_console_header")
        self.gridLayout.addWidget(self.label_console_header, 0, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label_messages_count = QtWidgets.QLabel(Form)
        self.label_messages_count.setObjectName("label_messages_count")
        self.horizontalLayout.addWidget(self.label_messages_count)
        self.input_messages_count = QtWidgets.QSpinBox(Form)
        self.input_messages_count.setReadOnly(True)
        self.input_messages_count.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)
        self.input_messages_count.setMaximum(2099999999)
        self.input_messages_count.setObjectName("input_messages_count")
        self.horizontalLayout.addWidget(self.input_messages_count)
        self.label_errors_count = QtWidgets.QLabel(Form)
        self.label_errors_count.setObjectName("label_errors_count")
        self.horizontalLayout.addWidget(self.label_errors_count)
        self.input_errors_count = QtWidgets.QSpinBox(Form)
        self.input_errors_count.setReadOnly(True)
        self.input_errors_count.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)
        self.input_errors_count.setMaximum(2099999999)
        self.input_errors_count.setObjectName("input_errors_count")
        self.horizontalLayout.addWidget(self.input_errors_count)
        self.button_pause_output = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_pause_output.sizePolicy().hasHeightForWidth())
        self.button_pause_output.setSizePolicy(sizePolicy)
        self.button_pause_output.setMinimumSize(QtCore.QSize(0, 0))
        self.button_pause_output.setMaximumSize(QtCore.QSize(16777215, 16777215))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/pause.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_pause_output.setIcon(icon)
        self.button_pause_output.setIconSize(QtCore.QSize(16, 16))
        self.button_pause_output.setObjectName("button_pause_output")
        self.horizontalLayout.addWidget(self.button_pause_output)
        self.button_resume_output = QtWidgets.QPushButton(Form)
        self.button_resume_output.setEnabled(False)
        self.button_resume_output.setMaximumSize(QtCore.QSize(16777215, 16777215))
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/icons/play.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_resume_output.setIcon(icon1)
        self.button_resume_output.setIconSize(QtCore.QSize(16, 16))
        self.button_resume_output.setObjectName("button_resume_output")
        self.horizontalLayout.addWidget(self.button_resume_output)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.button_clear_console = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_clear_console.sizePolicy().hasHeightForWidth())
        self.button_clear_console.setSizePolicy(sizePolicy)
        self.button_clear_console.setMinimumSize(QtCore.QSize(0, 0))
        self.button_clear_console.setMaximumSize(QtCore.QSize(16777215, 16777215))
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/icons/icons/cancel.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_clear_console.setIcon(icon2)
        self.button_clear_console.setIconSize(QtCore.QSize(16, 16))
        self.button_clear_console.setObjectName("button_clear_console")
        self.horizontalLayout.addWidget(self.button_clear_console)
        self.button_scroll_down = QtWidgets.QPushButton(Form)
        self.button_scroll_down.setMaximumSize(QtCore.QSize(16777215, 16777215))
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/icons/plus_double.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_scroll_down.setIcon(icon3)
        self.button_scroll_down.setIconSize(QtCore.QSize(16, 16))
        self.button_scroll_down.setObjectName("button_scroll_down")
        self.horizontalLayout.addWidget(self.button_scroll_down)
        self.button_scroll_down_2 = QtWidgets.QPushButton(Form)
        self.button_scroll_down_2.setMaximumSize(QtCore.QSize(16777215, 16777215))
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap(":/icons/icons/minus_double.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_scroll_down_2.setIcon(icon4)
        self.button_scroll_down_2.setIconSize(QtCore.QSize(16, 16))
        self.button_scroll_down_2.setObjectName("button_scroll_down_2")
        self.horizontalLayout.addWidget(self.button_scroll_down_2)
        self.gridLayout.addLayout(self.horizontalLayout, 4, 0, 1, 1)

        self.retranslateUi(Form)
        self.button_clear_console.clicked.connect(Form.clear_all)
        self.button_pause_output.clicked.connect(Form.pause_output)
        self.button_resume_output.clicked.connect(Form.resume_output)
        self.button_scroll_down_2.clicked.connect(Form.scroll_down)
        self.button_scroll_down.clicked.connect(Form.scroll_up)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label_console_header.setText(_translate("Form", "Standard output"))
        self.label_messages_count.setText(_translate("Form", "Messages:"))
        self.label_errors_count.setText(_translate("Form", "Errors:"))
        self.button_pause_output.setToolTip(_translate("Form", "Clear console output"))
        self.button_pause_output.setText(_translate("Form", " Pause output"))
        self.button_resume_output.setText(_translate("Form", " Resume output"))
        self.button_clear_console.setToolTip(_translate("Form", "Clear console output"))
        self.button_clear_console.setText(_translate("Form", " Clear all"))
        self.button_scroll_down.setText(_translate("Form", "Up"))
        self.button_scroll_down_2.setText(_translate("Form", "Down"))

from kamzik3 import resource_kamzik3_rc



import logging
import os
import time
from collections import OrderedDict
from copy import copy
from multiprocessing import Process
from threading import Thread

import numpy as np
import zmq

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsZmq import BalancedServer, Worker


class FileWorker(Worker, Process):

    def __init__(self, id, master_host, master_port, worker_host, worker_port=None, zmq_context=None):
        Process.__init__(self)
        Worker.__init__(self, id, master_host, master_port, zmq_context)
        if not hasattr(self, "logger"):
            self.logger = logging.getLogger("Device.{}".format(id))
        self.worker_host = worker_host
        self.worker_port = worker_port

    def _init_socket(self):
        if self.zmq_context is None:
            self.zmq_context = zmq.Context.instance()

        self.socket = self.zmq_context.socket(zmq.ROUTER)
        # Probe router ensures, that handshake is initiated after router is connected to other socket
        self.socket.probe_router = 1
        self.socket.setsockopt(zmq.IDENTITY, self.id)
        self.socket.setsockopt(zmq.LINGER, -1)
        if self.worker_port is None:
            self.worker_port = self.socket.bind_to_random_port("tcp://{}".format(self.worker_host))
        else:
            self.socket.bind("tcp://{}:{}".format(self.worker_host, self.worker_port))
        # Connect to Master host:port
        self.socket.connect("tcp://{}:{}".format(self.master_host, self.master_port))

    def run(self):
        self.running = True
        self._init_socket()
        # Wait for ACK from Master
        self.master_address, _, message = self.socket.recv_multipart()

        if message == MSG_ACK:
            # Run until we got MSG_TERMINATE from Master
            while self.running:
                current_task = self.socket.recv_multipart()
                if len(current_task) == 3 and current_task[2] == MSG_TERMINATE:
                    self.stop()
                elif current_task[4] == INSTRUCTION_PUSH:
                    self.task_push(current_task)
                elif current_task[4] in (INSTRUCTION_PULL, INSTRUCTION_SYNC):
                    self.task_pull(current_task)
                elif current_task[4] == INSTRUCTION_FORWARD:
                    self.task_forward(current_task)

    def task_forward(self, task_data):
        self.socket.send_multipart(
            [task_data[0], MSG_EMPTY, MSG_FORWARD, task_data[2], self.worker_host.encode(),
             str(self.worker_port).encode(), task_data[-1]])
        writer_address, _ = self.socket.recv_multipart()
        drive, source_file, remote_host, remote_port = task_data[5:9]
        self.logger.info("Forwarding file {}:{}".format(drive.decode(), source_file.decode()))
        # Connect to remote host
        self.socket.connect("tcp://{}:{}".format(remote_host.decode(), remote_port.decode()))
        remote_address, _, message = self.socket.recv_multipart()

        if message == MSG_ACK:
            # Request new Read Task
            self.socket.send_multipart([remote_address, MSG_EMPTY, INSTRUCTION_PUSH, drive, source_file])
            # Obtain Reader credentials
            _, _, reader_host, reader_port, reader_task = self.socket.recv_multipart()
            # Connect to Reader
            self.socket.connect("tcp://{}:{}".format(reader_host.decode(), reader_port.decode()))
            # Finishing handshake with Reader
            reader_address, _ = self.socket.recv_multipart()
            # Waiting for metadata from Reader
            file_metadata = self.socket.recv_multipart()
            # Send file metadata to writer
            self.socket.send_multipart([writer_address, MSG_EMPTY] + file_metadata[2:], copy=False)
            # Wait for ACK from Writer
            self.socket.recv_multipart()

            aborted = False
            progress_sync_threshold, sync_increment = 5, 5
            message = None
            while True:
                if message is None:
                    # Notify Reader, that we are ready to receive file chunk
                    self.socket.send_multipart([reader_address, MSG_EMPTY, WORKER_READY])
                    # Read file chunk from Reader
                    message = self.socket.recv_multipart()

                # Stop writing on EOF
                if len(message) == 2:
                    self.socket.send_multipart([reader_address, MSG_EMPTY, WORKER_DONE])
                    break
                elif message[0] == self.master_address and message[2] == MSG_ABORT:
                    # Wait for ACK from writer
                    self.socket.recv_multipart()
                    # Notify reader, that task was Aborted
                    self.socket.send_multipart([reader_address, MSG_EMPTY, MSG_ABORT])
                    # Wait for ACK from reader
                    self.socket.recv_multipart()
                    # Notify reader, that task was Aborted
                    self.socket.send_multipart([writer_address, MSG_EMPTY, MSG_ABORT])
                    # Wait for ACK from reader
                    self.socket.recv_multipart()
                    # Reply back to Master with ACK
                    self.socket.send_multipart([self.master_address, MSG_EMPTY, MSG_ACK])
                    aborted = True
                    break
                elif message[0] == reader_address and message[2] == MSG_ABORT:
                    # Notify Master, that task was aborted
                    self.socket.send_multipart([self.master_address, MSG_EMPTY, MSG_ABORT, task_data[-1]])
                    # Wait for ACK from Master
                    self.socket.recv_multipart()
                    # Notify reader, that task was Aborted
                    self.socket.send_multipart([writer_address, MSG_EMPTY, MSG_ABORT])
                    # Wait for ACK from reader
                    self.socket.recv_multipart()
                    # Reply back to Reader with ACK
                    self.socket.send_multipart([reader_address, MSG_EMPTY, MSG_ACK])
                    aborted = True
                    break
                elif message[0] == writer_address and message[2] == MSG_ABORT:
                    # Notify Master, that task was aborted
                    self.socket.send_multipart([self.master_address, MSG_EMPTY, MSG_ABORT, task_data[-1]])
                    # Wait for ACK from Master
                    self.socket.recv_multipart()
                    # Notify reader, that task was Aborted
                    self.socket.send_multipart([reader_address, MSG_EMPTY, MSG_ABORT])
                    # Wait for ACK from reader
                    self.socket.recv_multipart()
                    # Reply back to Reader with ACK
                    self.socket.send_multipart([writer_address, MSG_EMPTY, MSG_ACK])
                    aborted = True
                    break

                self.socket.send_multipart([writer_address, MSG_EMPTY] + message[2:], copy=False)
                # Wait for ACK from Writer
                ack = self.socket.recv_multipart()
                if ack[0] == writer_address and ack[2] == WORKER_READY:
                    # Check if we should sync progress with server
                    if float(message[3].decode()) >= progress_sync_threshold:
                        self.socket.send_multipart(
                            [self.master_address, MSG_EMPTY, MSG_PROGRESS, task_data[-1], message[3]])
                        progress_sync_threshold += sync_increment
                    message = None
                else:
                    message = ack

            if not aborted:
                message = self.socket.recv_multipart()
                if message[2] == MSG_ACK:
                    self.socket.send_multipart([writer_address, MSG_EMPTY], copy=False)
                    writer_addr, _, status = self.socket.recv_multipart()
                    if status == WORKER_DONE:
                        self.socket.send_multipart([writer_addr, MSG_EMPTY, MSG_ACK], copy=False)
                        self.socket.send_multipart([self.master_address, MSG_EMPTY, WORKER_DONE, task_data[-1]])

            # Close connection to Reader
            self.socket.disconnect("tcp://{}:{}".format(reader_host.decode(), reader_port.decode()))
            # Close connection to remote File server
            self.socket.disconnect("tcp://{}:{}".format(remote_host.decode(), remote_port.decode()))

    def task_push(self, task_data):
        # Reply to Master and forward reply to original client
        self.socket.send_multipart(
            [task_data[0], MSG_EMPTY, MSG_FORWARD, task_data[2], self.worker_host.encode(),
             str(self.worker_port).encode(), task_data[-1]])
        # Wait for writer and finish handshake
        writer_address, _ = self.socket.recv_multipart()
        source_file = task_data[6].decode()
        file_size = os.path.getsize(source_file)
        file_mod_time = os.path.getmtime(source_file)
        chunk_size = int(5e5)
        self.logger.info("Pushing file {}".format(source_file))
        # Send file metadata to writer
        self.socket.send_multipart(
            [writer_address, MSG_EMPTY, source_file.encode(), str(file_size).encode(), str(file_mod_time).encode(),
             str(chunk_size).encode()], copy=False)

        chunks_send = 0
        bytes_send = 0
        aborted = False
        progress_sync_threshold, sync_increment = 5, 5
        with open(source_file, "rb") as fp:
            while self.running:
                message = self.socket.recv_multipart()
                if message[0] == self.master_address and message[2] == MSG_ABORT:
                    # Wait until writer is ready
                    self.socket.recv_multipart()
                    # Notify writer that task is aborted
                    self.socket.send_multipart([writer_address, MSG_EMPTY, MSG_ABORT])
                    # Get ACK from writer
                    self.socket.recv_multipart()
                    # Reply back to master with ACK
                    self.socket.send_multipart([self.master_address, MSG_EMPTY, MSG_ACK])
                    aborted = True
                    break
                elif message[0] == writer_address and message[2] == MSG_ABORT:
                    # Notify Master, that task was aborted
                    self.socket.send_multipart([self.master_address, MSG_EMPTY, MSG_ABORT, task_data[-1]])
                    # Get ACK from Master
                    self.socket.recv_multipart()
                    # Reply back to writer, that task was aborted
                    self.socket.send_multipart([writer_address, MSG_EMPTY, MSG_ACK])
                    aborted = True
                    break
                # Read chunk from file
                data = fp.read(chunk_size)
                # No data = EOF
                if not data:
                    self.socket.send_multipart([writer_address, MSG_EMPTY], copy=False)
                    break
                chunks_send += 1
                bytes_send += len(data)
                # Calculate read progress
                progress = (bytes_send / file_size) * 100
                # Send data chunk to writer
                self.socket.send_multipart([writer_address, MSG_EMPTY, data, str(progress).encode()], copy=False)
                # Check if we should sync progress with server
                if progress >= progress_sync_threshold:
                    # Sync progress with server
                    self.socket.send_multipart(
                        [self.master_address, MSG_EMPTY, MSG_PROGRESS, task_data[-1], str(progress).encode()])
                    # Set next progress threshold
                    progress_sync_threshold += sync_increment

        # Push Task was finished successfully
        if not aborted:
            writer_addr, _, status = self.socket.recv_multipart()
            if status == WORKER_DONE:
                self.socket.send_multipart([writer_addr, MSG_EMPTY, MSG_ACK], copy=False)
                self.socket.send_multipart([self.master_address, MSG_EMPTY, WORKER_DONE, task_data[-1]])

    def task_pull(self, task_data):
        # If client and worker_id is equal, don't reply back to master
        if task_data[2] != self.id:
            # Reply to Master with info that should be forwarded to future Writer
            self.socket.send_multipart(
                [self.master_address, MSG_EMPTY, MSG_FORWARD, task_data[2], self.id, task_data[-1]])
        remote_host, remote_port, drive, source_file, dest_file, task_id = task_data[5:]
        self.logger.info(
            "Pulling file {}:{} and saving to {}".format(drive.decode(), source_file.decode(), dest_file.decode()))
        # Connect to remote File server
        self.socket.connect("tcp://{}:{}".format(remote_host.decode(), remote_port.decode()))
        # Get acknowledge response from remote File server
        remote_address, _, message = self.socket.recv_multipart()
        if message == MSG_ACK:
            # Request new Read Task
            self.socket.send_multipart([remote_address, MSG_EMPTY, INSTRUCTION_PUSH, drive, source_file])
            # Obtain Reader credentials
            _, _, reader_host, reader_port, reader_task = self.socket.recv_multipart()
            # Connect to Reader
            self.socket.connect("tcp://{}:{}".format(reader_host.decode(), reader_port.decode()))
            # Finishing handshake with Reader
            reader_address, _ = self.socket.recv_multipart()
            # Waiting for metadata from Reader
            _, _, source_file_path, file_size, file_mod_time, chunk_size = self.socket.recv_multipart()

            chunks_received = 0
            bytes_received = 0
            aborted = False
            progress_sync_threshold, sync_increment = 5, 5
            # Create or overwrite destination file
            with open(dest_file, "wb") as fp:
                # Notify reader that socket is ready to acquire file content
                while True:
                    # Notify Reader, that we are ready to receive file chunk
                    self.socket.send_multipart([reader_address, MSG_EMPTY, WORKER_READY])
                    # Read file chunk from Reader
                    message = self.socket.recv_multipart()
                    # Stop writing on EOF
                    if len(message) == 2:
                        self.socket.send_multipart([reader_address, MSG_EMPTY, WORKER_DONE])
                        break
                    elif message[0] == self.master_address and message[2] == MSG_ABORT:
                        # Notify reader, that task was Aborted
                        self.socket.send_multipart([reader_address, MSG_EMPTY, MSG_ABORT])
                        # Wait for ACK from reader
                        self.socket.recv_multipart()
                        # Reply back to Master with ACK
                        self.socket.send_multipart([self.master_address, MSG_EMPTY, MSG_ACK])
                        aborted = True
                        break
                    elif message[0] == reader_address and message[2] == MSG_ABORT:
                        # Notify Master, that task was aborted
                        self.socket.send_multipart([self.master_address, MSG_EMPTY, MSG_ABORT, task_data[-1]])
                        # Wait for ACK from Master
                        self.socket.recv_multipart()
                        # Reply back to Reader with ACK
                        self.socket.send_multipart([reader_address, MSG_EMPTY, MSG_ACK])
                        aborted = True
                        break
                    chunks_received += 1
                    bytes_received += len(message[2])
                    fp.write(message[2])
                    # Check if we should sync progress with server
                    if float(message[3].decode()) >= progress_sync_threshold:
                        self.socket.send_multipart([self.master_address, MSG_EMPTY, MSG_PROGRESS, task_id, message[3]])
                        progress_sync_threshold += sync_increment

            if not aborted:
                os.utime(dest_file, (float(file_mod_time.decode()), float(file_mod_time.decode())))
                message = self.socket.recv_multipart()
                if message[2] == MSG_ACK:
                    self.socket.send_multipart([self.master_address, MSG_EMPTY, WORKER_DONE, task_id, source_file_path])
            else:
                # Remove partial file
                os.remove(dest_file)

            # Close connection to Reader
            self.socket.disconnect("tcp://{}:{}".format(reader_host.decode(), reader_port.decode()))
            # Close connection to remote File server
            self.socket.disconnect("tcp://{}:{}".format(remote_host.decode(), remote_port.decode()))


class DeviceFileServer(Device, BalancedServer, Thread):
    max_task_history = 5

    def __init__(self, device_id, host, port, workers_count=5, worker_class=FileWorker, config=None):
        self.task_counter = (i for i in range(2 ** 63))
        Thread.__init__(self)
        BalancedServer.__init__(self, host, port, workers_count, worker_class)
        Device.__init__(self, device_id, config)
        self.frontend_connector = zmq.Context.instance().socket(zmq.REQ)
        self.frontend_connector.connect("tcp://{}:{}".format(host, port))
        self.backend_connector = zmq.Context.instance().socket(zmq.REQ)
        self.connect()

    def _init_workers(self):
        for i in range(self.workers_count):
            worker = self.worker_class("W_{}@{}:{}".format(i, self.host, self.port), self.host, self.port_backend,
                                       self.host)
            self.local_workers.append(worker)
            worker.start()

    def main_loop(self):
        """
        This is main loop of Fileserver.
        First connect backend_connector.
        Then poll for messages waiting for sockets to be processed.
        First check if there is any message waiting in backend, then check frontend socket.
        :return: None
        """
        self.backend_connector.connect("tcp://{}:{}".format(self.host, self.port_backend))
        while self.running:
            # Poll sockets with 10ms timeout
            sockets = dict(self.poller.poll(timeout=10))

            if self.socket_backend in sockets:
                # Handle worker activity on the backend
                request = self.socket_backend.recv_multipart()
                if len(request) == 2:
                    # Worker connected for the first time
                    self._worker_idle(request[0])
                    # Send ACK to worker
                    self.socket_backend.send_multipart([request[0], MSG_EMPTY, MSG_ACK])
                    continue

                worker_address, _ = request[:2]
                if request[2] == MSG_FORWARD:
                    # Forward reply from worker to original client
                    self.set_task_status(request[-1].decode(), STATUS_BUSY)
                    # Task was accepted
                    self.socket_frontend.send_multipart([request[3], MSG_EMPTY] + request[4:])
                elif request[2] == MSG_PROGRESS:
                    # Task progress was updated
                    self.set_task_progress(request[3].decode(), float(request[4].decode()))
                elif request[2] == WORKER_DONE:
                    task_id = request[3].decode()
                    if self.get_value([task_id, ATTR_TASK_INSTRUCTION]) == "Sync files":
                        # If task is to sync file continue with next file
                        if self._sync_next_file(task_id, request[0], request[4].decode()):
                            continue
                    # Task is done
                    self.set_task_status(task_id, STATUS_IDLE)
                    self.set_value([task_id, ATTR_FINISHED_AT], time.time())
                    self._worker_idle(request[0])
                elif request[2] == MSG_ABORT:
                    self.socket_backend.send_multipart([request[0], MSG_EMPTY, MSG_ACK])
                    worker_address = self.get_value([request[3].decode(), ATTR_WORKER])
                    if worker_address.encode() != request[0]:
                        self.socket_backend.send_multipart([worker_address.encode(), MSG_EMPTY, MSG_ABORT, request[3]])
                    self.abort_task(request[3].decode())
                    self._worker_idle(worker_address.encode())
                elif request[2] == MSG_ACK:
                    pass

            if self.socket_frontend in sockets:
                # Get next client request, route to last-used worker
                request = self.socket_frontend.recv_multipart()
                if len(request) == 2:
                    self.socket_frontend.send_multipart([request[0], MSG_EMPTY, MSG_ACK])
                    continue
                if request[2] == INSTRUCTION_PUSH and not self.on_fs(request[3].decode(), request[4].decode()):
                    request[2] = INSTRUCTION_FORWARD
                    source = self.get_value(ATTR_DRIVES)[request[3].decode()]["source"]
                    request += [source.get_value(ATTR_HOST).encode(), str(source.get_value(ATTR_PORT)).encode()]

                worker = self.idle_workers.pop(0)
                task_id = self.add_task(worker.decode(), request)

                if request[2] == INSTRUCTION_SYNC:
                    self.socket_frontend.send_multipart([request[0], MSG_EMPTY, worker, task_id.encode()])
                    if not self._sync_next_file(task_id, worker):
                        self.set_task_status(task_id, STATUS_IDLE)
                        self.idle_workers.insert(0, worker)
                else:
                    self.socket_backend.send_multipart([worker, MSG_EMPTY] + request + [task_id.encode()])
                self._check_idle_workers()

    def _worker_idle(self, worker_id):
        """
        We have free worker, so push it back to idle_workers list.
        :param worker_id: str
        :return: None
        """
        self.idle_workers.append(worker_id)
        self._check_idle_workers()

    def _check_idle_workers(self):
        """
        Check if worker is IDLE.
        If so, make sure that fronted is reading requests.
        Otherwise remove fronted from poller.
        :return: None
        """
        if self.idle_workers:
            # Poll for clients now that a worker is available
            self.poller.register(self.socket_frontend, zmq.POLLIN)
        else:
            # Unregister fronted socket from polling, no free worker available
            self.poller.unregister(self.socket_frontend)

    def _sync_next_file(self, task_id, worker_id, synced_file=None):
        """
        This is specific to sync request.
        When multiple files are requested, sync Task uses one worker to sync files one by one.
        At each call, function determine which files are already synced and which still remains.
        If next file needs to be synced True is returned False otherwise.
        :param task_id: str
        :param worker_id: str
        :return: bool
        """
        self.set_value([task_id, ATTR_STATUS], STATUS_BUSY)
        requested_files = self.get_value([task_id, ATTR_REQUESTED_FILES])
        synced_files = OrderedDict()

        for drive, path in self.get_value([task_id, ATTR_REQUESTED_FILES]):
            input_dir, file_name = os.path.split(path)
            synced, local_path = self.get_local_path(drive, input_dir, file_name)
            # File was downloaded
            # What can happened is that file updated while checked for sync again
            # To avoid that I always assume that file is in sync when downloaded and existing on local FS
            if synced or path == synced_file:
                synced_files[(drive, path)] = local_path

        self.set_value([task_id, ATTR_SYNCED_FILES], synced_files)

        remaining_files = set(requested_files) - set(synced_files.keys())
        if len(remaining_files) > 0:
            drive, path = remaining_files.pop()
            input_dir, file_path = os.path.split(path)
            drive_meta = self.get_value(ATTR_DRIVES)[drive]
            source = drive_meta.get("source")
            cached, cache_path = self.is_cached(drive, input_dir, file_path)
            self.logger.info("Syncing file {}:{}".format(drive, path))
            self.socket_backend.send_multipart(
                [worker_id, MSG_EMPTY, worker_id, MSG_EMPTY, INSTRUCTION_SYNC, source.get_value(ATTR_HOST).encode(),
                 str(source.get_value(ATTR_PORT)).encode(),
                 drive.encode(), path.encode(), cache_path.encode(), task_id.encode()])
            return True
        else:
            return False

    def clean_task_history(self):
        """
        Clear all IDLE tasks to maximum set by max_task_history
        :return: None
        """
        task_count = 0
        for group in reversed(copy(list(self.attributes.keys()))):
            if group.startswith("T_"):
                if self.get_value([group, ATTR_STATUS]) == STATUS_IDLE and task_count >= self.max_task_history - 1:
                    self.remove_task(group)
                else:
                    task_count += 1

    def add_task(self, worker_id, request):
        """
        Add new task for Fileserver. 
        Each Task is represented by request which is then given to free worker.
        After new task is added task_id is returned back.
        :param worker_id: str
        :param request: list
        :return: str
        """
        self.clean_task_history()
        instruction = request[2]
        task_id = "T_{}".format(next(self.task_counter))
        self.logger.info("Task {} added".format(task_id))

        self.create_attribute(ATTR_STATUS, task_id, default_value=STATUS_CONFIGURING, readonly=True)
        self.create_attribute(ATTR_STARTED_AT, task_id, default_value=time.time(), unit="sec", readonly=True,
                              default_type=np.float)
        self.create_attribute(ATTR_FINISHED_AT, task_id, unit="sec", readonly=True, default_type=np.float)
        self.create_attribute(ATTR_PROGRESS, task_id, default_value=0, unit="%", readonly=True, default_type=np.float,
                              decimals=2)
        self.create_attribute(ATTR_WORKER, task_id, default_value=worker_id, readonly=True)
        if instruction == INSTRUCTION_PULL:
            self.create_attribute(ATTR_TASK_INSTRUCTION, task_id, default_value="Pull file", readonly=True)
            self.create_attribute(ATTR_ADDRESS, task_id,
                                  default_value="tcp://{}:{}".format(request[3].decode(), request[4].decode()),
                                  readonly=True)
            self.create_attribute(ATTR_SOURCE_FILE, task_id, default_value=request[5].decode(), readonly=True)
            self.create_attribute(ATTR_TARGET_FILE, task_id, default_value=request[6].decode(), readonly=True)
        elif instruction == INSTRUCTION_PUSH:
            self.create_attribute(ATTR_TASK_INSTRUCTION, task_id, default_value="Push file", readonly=True)
            self.create_attribute(ATTR_SOURCE_FILE, task_id, default_value=request[4].decode(), readonly=True)
        elif instruction == INSTRUCTION_FORWARD:
            self.create_attribute(ATTR_TASK_INSTRUCTION, task_id, default_value="Forward file", readonly=True)
            self.create_attribute(ATTR_SOURCE_FILE, task_id, default_value=request[3].decode(), readonly=True)
        elif instruction == INSTRUCTION_SYNC:
            number_of_files = int(request[3].decode())
            self.create_attribute(ATTR_TASK_INSTRUCTION, task_id, default_value="Sync files", readonly=True)
            offset = 4
            syncing_files = []
            for i in range(number_of_files):
                drive, path = request[offset:offset + 2]
                offset += 2
                syncing_files.append((drive.decode(), path.decode()))
            self.create_attribute(ATTR_REQUESTED_FILES, task_id, default_value=syncing_files, readonly=True)
            self.create_attribute(ATTR_SYNCED_FILES, task_id, default_value=[], readonly=True)
        self.create_attribute(ATTR_ABORTED, task_id, default_value=False, default_type=bool, readonly=True)
        return task_id

    def remove_task(self, task_id):
        """
        Remove task from list of all tasks.
        :param task_id: str
        :return: None
        """
        if task_id in self.attributes:
            self.logger.info("Task {} removed".format(task_id))
            self.delete_attribute_group(task_id)

    def set_task_progress(self, task_id, progress):
        """
        Set new task progress.
        If syncing multiple files, calculate overall progress based on number of files.
        :param task_id: str
        :param progress: float
        :return: None
        """
        if self.get_value([task_id, ATTR_TASK_INSTRUCTION]) == "Sync files":
            requested_files = len(self.get_value([task_id, ATTR_REQUESTED_FILES]))
            synced_files = len(self.get_value([task_id, ATTR_SYNCED_FILES]).keys())
            sync_file_progress_part = 100 / requested_files
            progress = sync_file_progress_part * synced_files + (sync_file_progress_part * (progress / 100))

        self.set_value([task_id, ATTR_PROGRESS], progress)

    def set_task_status(self, task_id, status):
        """
        Set new task status.
        :param task_id: str
        :param status: str
        :return: None
        """
        self.set_value([task_id, ATTR_STATUS], status)

    def abort_task(self, task_id):
        """
        Task was aborted for any reason.
        :param task_id:
        :return: None
        """
        self.logger.info("Task {} aborted".format(task_id))
        self.set_value([task_id, ATTR_ABORTED], True)
        self.set_task_status(task_id, STATUS_IDLE)

    @expose_method({"task_id": "Task id"})
    def stop_task(self, task_id):
        """
        Initiate task to be stopped, based on task_id.
        :param task_id: str
        :return: None
        """
        if task_id not in self.attributes:
            raise DeviceError("Task '{}' does not exists".format(task_id))
        elif self.get_value([task_id, ATTR_STATUS]) != STATUS_BUSY:
            raise DeviceError("Task '{}' is not running".format(task_id))
        self.backend_connector.send_multipart([MSG_ABORT, task_id.encode()])
        self.backend_connector.recv_multipart()

    @expose_method()
    def stop_all_tasks(self):
        """
        Stop all running Tasks
        :return: None
        """
        self.logger.info("Stop all tasks")
        for group in self.attributes.keys():
            if group == "General" or self.get_value([group, ATTR_STATUS]) != STATUS_BUSY:
                continue
            self.stop_task(group)

    def stop(self):
        """
        Stop Fileserver
        :return:
        """
        for worker in self.local_workers:
            self.socket_backend.send_multipart([worker.id, MSG_EMPTY, MSG_TERMINATE])

        self.local_workers = []
        self.running = False

    def close(self):
        self.stop()
        Device.close(self)



import numpy as np
import serial
import time
from bidict import frozenbidict
from threading import Thread

import kamzik3
from kamzik3 import DeviceError, units
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.devices.deviceChannel import DeviceChannel
from kamzik3.devices.deviceGalil import DeviceGalil
from kamzik3.devices.devicePort import DevicePort
from kamzik3.macro.step import StepDeviceAttributeNumerical
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import device_units

MOTOR_TYPE_PWM_SIGN_SERVO = u"PMW sign servo (1.5)"
MOTOR_TYPE_PWM_SIGN_SERVO_REVERSED_POLARITY = u"PMW servo reversed polarity (-1.5)"
MOTOR_TYPE_SERVO = u"Servo (1)"
MOTOR_TYPE_SERVO_REVERSED_POLARITY = u"Servo reversed polarity (-1)"
MOTOR_TYPE_STEP_ACTIVE_HIGH_REVERSED = u"Step active high reversed (-2.5)"
MOTOR_TYPE_STEP_ACTIVE_HIGH_STEP = u"Step active high step (-2)"
MOTOR_TYPE_STEP_ACTIVE_LOW_REVERSED = u"Step active low reversed (2.5)"
MOTOR_TYPE_STEP_ACTIVE_LOW_STEP = u"Step active low step (2)"

MOTOR_TYPES = frozenbidict({
    MOTOR_TYPE_SERVO: 1,
    MOTOR_TYPE_SERVO_REVERSED_POLARITY: -1,
    MOTOR_TYPE_PWM_SIGN_SERVO: 1.5,
    MOTOR_TYPE_PWM_SIGN_SERVO_REVERSED_POLARITY: -1.5,
    MOTOR_TYPE_STEP_ACTIVE_LOW_STEP: 2,
    MOTOR_TYPE_STEP_ACTIVE_HIGH_STEP: -2,
    MOTOR_TYPE_STEP_ACTIVE_LOW_REVERSED: 2.5,
    MOTOR_TYPE_STEP_ACTIVE_HIGH_REVERSED: -2.5,
})
LIMIT_SWITCH_LEVELS = frozenbidict({
    u"High": 1,
    u"Low": -1
})
HOMING_DIRECTION = frozenbidict({
    u"Forward": 1,
    u"Backward": -1
})
HOMING_METHODS = frozenbidict({
    u"Command based": 0,
    u"Limit based": 1
})
LIMIT_SWITCHES_MODE = frozenbidict({
    u"Both limit switches are enabled": 0,
    u"Forward limit switch disabled": 1,
    u"Reverse limit switch disabled": 2,
    u"Both limit switch disabled": 3
})
FLAG_FORWARD_LIMIT_INACTIVE = u"Forward limit inactive flag"
FLAG_HOME_SWITCH = u"Home switch flag"
FLAG_MOTOR_MOVING = u"Motor moving flag"
FLAG_MOTOR_OFF = u"Motor Off flag"
FLAG_POSITION_EXCEEDS_ERROR = u"Position error flag"
FLAG_POSITION_LATCH = u"Position latch occured flag"
FLAG_REFERENCING = u"Referencing flag"
FLAG_REVERSE_LIMIT_INACTIVE = u"Reversed limit inactive flag"


class DeviceGalilMotor(DeviceGalil):
    acronym_letters = u"ABCDEFGHIJKLMNOPQRSTUVXYZ"
    push_commands_max = 1
    response_timeout = 3000

    def _init_attributes(self):
        Device._init_attributes(self)

        self.create_attribute(ATTR_CHANNELS, readonly=True, default_type=np.uint8, min_value=0, max_value=256)
        self.create_attribute(ATTR_LIMIT_SWITCH_ACTIVE_LEVEL, default_type=LIMIT_SWITCH_LEVELS.keys(),
                              set_function=self.set_limit_switch_active_level)
        self.create_attribute(ATTR_HOMING_DIRECTION, default_type=HOMING_DIRECTION.keys(),
                              set_function=self.set_homing_direction)
        self.create_attribute(ATTR_LATCH_INPUT_LEVEL, default_type=LIMIT_SWITCH_LEVELS.keys(),
                              set_function=self.set_latch_input_level)

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.command(u"EO 0;")
        # Get number of channels
        self.command(u"QZ;")
        self.get_motor_config(callback=_finish_configuration)

    def start_polling(self):
        Device.start_polling(self)
        self.poll_command(u"TD;", 300)
        self.poll_command(u"TS;", 300)

    def set_limit_switch_active_level(self, value, callback=None):
        self.command(u"CN {};".format(LIMIT_SWITCH_LEVELS.get(value)), callback=callback)

    def set_homing_direction(self, value, callback=None):
        self.command(u"CN ,{};".format(HOMING_DIRECTION.get(value)), callback=callback)

    def set_latch_input_level(self, value, callback=None):
        self.command(u"CN ,,{};".format(LIMIT_SWITCH_LEVELS.get(value)), callback=callback)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = super(DeviceGalil, self).handle_readout(readout_buffer)
        # Strip white characters from output
        output = output.strip()
        if output and output[0] == "?":
            self.handle_command_error(command, output)
            self.get_error_code()
            return False
        elif command == u"MG _CN0, _CN1, _CN2, _CN3, _CN4;":
            ls_level, home_direction, latch_input_level, selective_abort, no_termniation_flag = [int(float(v)) for v in
                                                                                                 output.split(" ")]
            self.set_raw_value(ATTR_LIMIT_SWITCH_ACTIVE_LEVEL, LIMIT_SWITCH_LEVELS.inv.get(ls_level))
            self.set_raw_value(ATTR_HOMING_DIRECTION, HOMING_DIRECTION.inv.get(home_direction))
            self.set_raw_value(ATTR_LATCH_INPUT_LEVEL, LIMIT_SWITCH_LEVELS.inv.get(latch_input_level))
        elif command == u"QZ;":
            parsed_output = [int(float(v)) for v in output.split(", ")]
            self.set_raw_value(ATTR_CHANNELS, parsed_output[0])
        # elif command == u"MG _QZ0;":
        #     self.set_raw_value(ATTR_CHANNELS, int(float(output)))
        elif command == u"TC 1;":
            readout_buffer = output
            self.handle_command_error(command, readout_buffer)
        elif command == u"TS;":
            for channel, status in enumerate((int(float(v)) for v in output.split(", "))):
                for bit, flag in enumerate((FLAG_POSITION_LATCH, FLAG_HOME_SWITCH, FLAG_REVERSE_LIMIT_INACTIVE,
                                            FLAG_FORWARD_LIMIT_INACTIVE, None,
                                            FLAG_MOTOR_OFF, FLAG_POSITION_EXCEEDS_ERROR, FLAG_MOTOR_MOVING)):
                    if flag is None: continue
                    state = bool(status & (1 << bit))
                    self.notify((channel, flag), state)
                    if flag == FLAG_MOTOR_MOVING:
                        self.notify((channel, ATTR_STATUS), STATUS_BUSY if state else STATUS_IDLE)
        elif command == u"TD;":
            for channel, position in enumerate((int(float(v)) for v in output.split(", "))):
                self.notify((channel, ATTR_POSITION), position)
        elif u"MG _AG" in command:
            channel = self.acronym_letters.find(command[-2])
            self.notify((channel, ATTR_AMPLIFIER_GAIN), int(float(output)))
        elif u"MG _MT" in command:
            channel = self.acronym_letters.find(command[-2])
            motor_type = MOTOR_TYPES.inv.get(float(output))
            self.notify((channel, ATTR_MOTOR_TYPE), motor_type)
        elif u"MG _YA" in command:
            channel = self.acronym_letters.find(command[-2])
            self.notify((channel, ATTR_STEP_DRIVE_RESOLUTION), int(float(output)))
        elif u"MG _SP" in command:
            channel = self.acronym_letters.find(command[-2])
            self.notify((channel, ATTR_VELOCITY), int(float(output)))
        elif u"MG _LC" in command:
            channel = self.acronym_letters.find(command[-2])
            self.notify((channel, ATTR_LOW_CURRENT_STEPPER_MODE), int(float(output)))
        elif u"MG _LD" in command:
            channel = self.acronym_letters.find(command[-2])
            limit_mode = LIMIT_SWITCHES_MODE.inv.get(float(output))
            self.notify((channel, ATTR_LIMIT_SWITCHES_MODE), limit_mode)
        elif u"MG _AC" in command:
            channel = self.acronym_letters.find(command[-2])
            self.notify((channel, ATTR_ACCELERATION), int(float(output)))
        elif u"MG _DC" in command:
            channel = self.acronym_letters.find(command[-2])
            self.notify((channel, ATTR_DECELERATION), int(float(output)))
        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(self.device_id, [command, readout_buffer], token)

            for observer in self._observers[:]:
                if isinstance(observer, DeviceGalilMotorChannel):
                    kamzik3.session.publisher.push_message(observer.device_id, [command, readout_buffer], token)

    def get_motor_config(self, callback=None):
        self.command(u"MG _CN0, _CN1, _CN2, _CN3, _CN4;", callback)


class DeviceGalilMotorPort(DevicePort, DeviceGalilMotor):
    terminator = DeviceGalil.terminator
    response_timeout = 3000

    def __init__(self, port, baud_rate=19200, parity=serial.PARITY_NONE, stop_bits=serial.STOPBITS_ONE,
                 byte_size=serial.EIGHTBITS, device_id=None, config=None):
        super(DeviceGalilMotorPort, self).__init__(port, baud_rate, parity, stop_bits, byte_size, device_id, config)

    def handshake(self):
        self.logger.info(u"Handshake initiated")
        self.push(b"MG 1;")
        time.sleep(0.02)  # Give some time for devices to reply
        if self.serial_port.in_waiting != 0:
            if self.read_all().strip() in (u"1.0000\r\n:", u"MG 1; 1.0000\r\n:"):
                return True
        return False


class DeviceGalilMotorChannel(DeviceChannel):

    def __init__(self, device, channel, device_id=None, config=None):
        self.acronym = DeviceGalilMotor.acronym_letters[channel]
        self.backlash_cleared_flag = True
        self.final_position_setpoint = None
        DeviceChannel.__init__(self, device, channel, device_id, config)

    def _init_attributes(self):
        super(DeviceGalilMotorChannel, self)._init_attributes()

        self.create_attribute(ATTR_ACRONYM, default_value=self.acronym, readonly=True)
        self.create_attribute(ATTR_AMPLIFIER_GAIN, default_type=np.uint8, min_value=0, max_value=3,
                              set_function=self.set_amplifier_gain)
        self.create_attribute(ATTR_MOTOR_TYPE, default_type=MOTOR_TYPES.keys(), set_function=self.set_motor_type)
        self.create_attribute(ATTR_VELOCITY, default_type=np.float, min_value=-15e6, max_value=15e6, unit=u"steps/sec",
                              set_function=self.set_velocity, decimals=3)
        self.create_attribute(ATTR_STEP_DRIVE_RESOLUTION, default_type=np.uint64, max_value=15e6, unit=u"ustep",
                              set_function=self.set_step_drive_resolution)
        self.create_attribute(ATTR_BACKLASH_DISTANCE, default_value=0, default_type=np.float64, readonly=False,
                              decimals=3)
        self.create_attribute(FLAG_FORWARD_LIMIT_INACTIVE, default_type=np.bool, readonly=True)
        self.create_attribute(FLAG_HOME_SWITCH, default_type=np.bool, readonly=True)
        self.create_attribute(FLAG_MOTOR_MOVING, default_type=np.bool, readonly=True)
        self.create_attribute(FLAG_MOTOR_OFF, default_type=np.bool, readonly=True)
        self.create_attribute(FLAG_POSITION_EXCEEDS_ERROR, default_type=np.bool, readonly=True)
        self.create_attribute(FLAG_POSITION_LATCH, default_type=np.bool, readonly=True)
        self.create_attribute(FLAG_REFERENCING, default_type=np.bool, readonly=True)
        self.create_attribute(FLAG_REVERSE_LIMIT_INACTIVE, default_type=np.bool, readonly=True)
        self.create_attribute(ATTR_LOW_CURRENT_STEPPER_MODE, default_type=np.uint32, max_value=32767,
                              set_function=self.set_low_current_stepper_mode)
        self.create_attribute(ATTR_ACCELERATION, default_type=np.float, min_value=1024, max_value=1073740800,
                              set_function=self.set_acceleration, unit="steps/sec^2", decimals=3)
        self.create_attribute(ATTR_DECELERATION, default_type=np.float, min_value=1024, max_value=1073740800,
                              set_function=self.set_deceleration, unit="steps/sec^2", decimals=3)
        self.create_attribute(ATTR_POSITION, default_type=np.float64, unit=u"counts", set_function=self._move_absolute,
                              set_value_when_set_function=False)
        self.create_attribute(ATTR_HOMING_METHOD, default_value=u"Limit based", default_type=HOMING_METHODS.keys())
        self.create_attribute(ATTR_LIMIT_SWITCHES_MODE, default_type=LIMIT_SWITCHES_MODE.keys(),
                              set_function=self.set_limit_switch_mode)
        self.create_attribute(ATTR_POSITION_LOWER_LIMIT, default_type=np.float64, unit=u"counts",
                              set_function=self.set_lower_position_limit)
        self.create_attribute(ATTR_POSITION_UPPER_LIMIT, default_type=np.float64, unit=u"counts",
                              set_function=self.set_upper_position_limit)
        self.create_attribute(ATTR_FAST_REFERENCE_VELOCITY, default_value=5, default_type=np.float, min_value=-15e6,
                              max_value=15e6, unit=u"steps/sec")
        self.create_attribute(ATTR_SLOW_REFERENCE_VELOCITY, default_value=1, default_type=np.float, min_value=-15e6,
                              max_value=15e6, unit=u"steps/sec")
        self.create_attribute(ATTR_SAFE_DIRECTION, default_value=u"Forward", default_type=HOMING_DIRECTION.keys())
        self.create_attribute(ATTR_REFERENCED, default_value=False, default_type=np.bool, readonly=True)
        self.create_attribute(ATTR_REFERENCE_STEPS_BACK, default_value=40, default_type=np.uint32, min_value=1,
                              unit=u"steps")
        self.create_attribute(ATTR_STEPS_PER_REVOLUTION, default_value=1, default_type=np.uint32)

    def handle_configuration(self):
        if self.configured:
            return

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.connected = True
        self.get_motor_type()
        self.get_step_drive_resolution()
        self.get_velocity()
        self.get_low_current_stepper_mode()
        self.get_acceleration()
        self.get_deceleration()
        self.get_limit_switch_mode()
        self.get_amplifier_gain(callback=_finish_configuration)

    def subject_update(self, key, value, subject):
        super(DeviceGalilMotorChannel, self).subject_update(key, value, subject)

        if self.connected and isinstance(key, tuple) and key[0] == self.channel:
            attribute = key[1]

            if attribute == ATTR_VELOCITY:
                value /= self.get_raw_value(ATTR_STEP_DRIVE_RESOLUTION)
            elif attribute == ATTR_ACCELERATION:
                value /= self.get_raw_value(ATTR_STEP_DRIVE_RESOLUTION)
            elif attribute == ATTR_DECELERATION:
                value /= self.get_raw_value(ATTR_STEP_DRIVE_RESOLUTION)
            elif attribute == ATTR_STATUS:
                if self.get_value(FLAG_REFERENCING):
                    value = STATUS_BUSY
                if value == STATUS_IDLE and not self.get_value(
                        FLAG_MOTOR_OFF) and self.get_value(ATTR_LOW_CURRENT_STEPPER_MODE) == 0:
                    self.servo_off()

            self.set_raw_value(ATTR_LATENCY, self.device.get_value(ATTR_LATENCY))
            self.set_raw_value(attribute, value)

    def _move_absolute(self, position):
        self.servo_on()
        position_diff = self[ATTR_POSITION][SETPOINT] - self[ATTR_POSITION][VALUE]
        backlash_distance = self.get_value(ATTR_BACKLASH_DISTANCE)
        if self.backlash_cleared_flag and backlash_distance != 0 and np.sign(backlash_distance) == np.sign(
                position_diff):
            backlash_macro = Thread(target=self._move_backlash, args=([self[ATTR_POSITION][SETPOINT]]))
            backlash_macro.start()
            return

        self.command(u"PA{}={:.10f};".format(self.acronym, position))
        self.begin_motion()

    def _move_backlash(self, new_setpoint):
        self.backlash_cleared_flag = False
        backlash_distance = self.get_value(ATTR_BACKLASH_DISTANCE)
        min_t, max_t = self[ATTR_POSITION].negative_tolerance(), self[ATTR_POSITION].positive_tolerance()
        device_unit = self.get_attribute(ATTR_POSITION)[UNIT]
        setpoint_one = units.Quantity(new_setpoint + backlash_distance, device_unit)
        setpoint_two = units.Quantity(new_setpoint, device_unit)
        step_one = StepDeviceAttributeNumerical("One", self.device_id, ATTR_POSITION, setpoint_one, min_t, max_t)
        step_two = StepDeviceAttributeNumerical("Two", self.device_id, ATTR_POSITION, setpoint_two, min_t, max_t)
        step_one.start()
        step_two.start()
        self.backlash_cleared_flag = True

    @expose_method({"steps": "Number of steps"})
    def move_steps(self, steps):
        resolution = self.get_value(ATTR_STEP_DRIVE_RESOLUTION)
        steps = int(steps) * resolution

        self.servo_on()
        self.command(u"PR{}={};".format(self.acronym, steps))
        self.begin_motion()

    @expose_method({"position": ATTR_POSITION})
    def move_relative(self, position):
        position = device_units(self, ATTR_POSITION, position)
        self.logger.info(u"Move motor channel by relative position {:~}".format(position))
        return self.set_value(ATTR_POSITION, self[ATTR_POSITION][VALUE] + position.m)

    @expose_method({"position": ATTR_POSITION})
    def move_absolute(self, position):
        position = device_units(self, ATTR_POSITION, position)
        self.logger.info(u"Move motor channel to absolute position {:~}".format(position))
        return self.set_value(ATTR_POSITION, position.m)

    def set_limit_switch_mode(self, value, callback=None):
        return self.command(u"LD{}={};".format(self.acronym, LIMIT_SWITCHES_MODE.get(value)), callback, with_token=True)

    def set_acceleration(self, value, callback=None):
        value = value * self.get_value(ATTR_STEP_DRIVE_RESOLUTION)
        return self.command(u"AC{}={};".format(self.acronym, int(value)), callback, with_token=True)

    def set_deceleration(self, value, callback=None):
        value = value * self.get_value(ATTR_STEP_DRIVE_RESOLUTION)
        return self.command(u"DC{}={};".format(self.acronym, int(value)), callback, with_token=True)

    def set_motor_type(self, motor_type, callback=None):
        return self.command(u"MT{}={};".format(self.acronym, MOTOR_TYPES.get(motor_type)), callback, with_token=True)

    def set_amplifier_gain(self, value, callback=None):
        return self.command(u"AG{}={};".format(self.acronym, int(value)), callback, with_token=True)

    def set_step_drive_resolution(self, value, callback=None):
        return self.command(u"YA{}={};".format(self.acronym, int(value)), callback, with_token=True)

    def set_velocity(self, value, callback=None):
        velocity = value * self.get_value(ATTR_STEP_DRIVE_RESOLUTION)
        return self.command(u"SP{}={};".format(self.acronym, int(velocity)), callback, with_token=True)

    def set_low_current_stepper_mode(self, value, callback=None):
        return self.command(u"LC{}={};".format(self.acronym, int(value)), callback, with_token=True)

    def set_lower_position_limit(self, value, callback=None):
        self.set_value(ATTR_POSITION_LOWER_LIMIT, value)
        return self.set_attribute([ATTR_POSITION, MIN], value, callback)

    def set_upper_position_limit(self, value, callback=None):
        self.set_value(ATTR_POSITION_UPPER_LIMIT, value)
        return self.set_attribute([ATTR_POSITION, MAX], value, callback)

    def get_motor_type(self, callback=None):
        self.command(u"MG _MT{};".format(self.acronym), callback)

    def get_amplifier_gain(self, callback=None):
        self.command(u"MG _AG{};".format(self.acronym), callback)

    def get_step_drive_resolution(self, callback=None):
        self.command(u"MG _YA{};".format(self.acronym), callback)

    def get_velocity(self, callback=None):
        self.command(u"MG _SP{};".format(self.acronym), callback)

    def get_position(self, callback=None):
        self.command(u"MG _TD{};".format(self.acronym), callback)

    def get_acceleration(self, callback=None):
        self.command(u"MG _AC{};".format(self.acronym), callback)

    def get_deceleration(self, callback=None):
        self.command(u"MG _DC{};".format(self.acronym), callback)

    def get_low_current_stepper_mode(self, callback=None):
        self.command(u"MG _LC{};".format(self.acronym), callback)

    def get_limit_switch_mode(self, callback=None):
        self.command(u"MG _LD{};".format(self.acronym), callback)

    def get_status(self, callback=None):
        self.command(u"MG _TS{};".format(self.acronym), callback)

    def begin_motion(self):
        self.command(u"BG{};".format(self.acronym))

    def servo_on(self):
        self.command(u"SH{};".format(self.acronym))

    def servo_off(self):
        self.command(u"MO{};".format(self.acronym))

    def _limit_based_reference(self):
        self.set_attribute([FLAG_REFERENCING, VALUE], True)
        self.set_attribute([ATTR_REFERENCED, VALUE], False)
        direction = HOMING_DIRECTION[self.get_value(ATTR_SAFE_DIRECTION)]
        fast_reference_velocity = self.get_value(ATTR_FAST_REFERENCE_VELOCITY)
        slow_reference_velocity = self.get_value(ATTR_SLOW_REFERENCE_VELOCITY)
        current_velocity = self.get_value(ATTR_VELOCITY)
        steps_back = self.get_value(ATTR_REFERENCE_STEPS_BACK)
        if direction == 1:
            check_limit = FLAG_FORWARD_LIMIT_INACTIVE
        else:
            check_limit = FLAG_REVERSE_LIMIT_INACTIVE

        def _reference_motion():
            # Jog fast until hard stop
            self.jog(fast_reference_velocity * direction)
            self._wait_until_stopped()
            # Check if we are at the limit
            if self.get_value(check_limit) == 0:
                # Move away from hard stop by 40 steps
                self.move_steps(steps_back * -direction)
                self._wait_until_stopped()
                # Jog slow until hard stop again
                self.jog(slow_reference_velocity * direction)
                self._wait_until_stopped()
                # Move away from hard stop by 40 steps
                self.move_steps(steps_back * -direction)
                self._wait_until_stopped()
                self.reset_position(0)
                self.set_velocity(current_velocity)
                self.set_attribute([FLAG_REFERENCING, VALUE], False)
                self.set_attribute([ATTR_REFERENCED, VALUE], True)
            else:
                self.set_attribute([FLAG_REFERENCING, VALUE], False)
                raise DeviceError(u"Referencing failed")

        Thread(target=_reference_motion).start()
        self.logger.debug(u"Start referencing in {} direction".format(self.get_value(ATTR_SAFE_DIRECTION)))

    def _wait_until_stopped(self):
        self.command(u"TS;")
        time.sleep(0.2)
        while self.get_value(FLAG_MOTOR_MOVING):
            time.sleep(0.1)
        return True

    @expose_method()
    def find_reference(self):
        homnig_method = self.get_value(ATTR_HOMING_METHOD)
        if homnig_method == u"Command based":
            self.servo_on()
            self.command(u"HM{};".format(self.acronym))
            self.begin_motion()
        elif homnig_method == u"Limit based":
            self._limit_based_reference()

    @expose_method({"position": ATTR_POSITION})
    def reset_position(self, position=0):
        position = device_units(self, ATTR_POSITION, position)
        self.logger.info(u"Reset motor position to {:~}".format(position))
        position = self[ATTR_POSITION].remove_offset_factor(position.m)
        self.command(u"DP{}={};".format(self.acronym, position))

    @expose_method({"velocity": ATTR_VELOCITY})
    def jog(self, velocity):
        self.logger.info(u"Start motor jogging")
        velocity = device_units(self, ATTR_VELOCITY, velocity)
        velocity = self[ATTR_POSITION].remove_offset_factor(velocity)
        velocity = float(velocity.m) * self.get_value(ATTR_STEP_DRIVE_RESOLUTION)
        self.servo_on()
        self.command(u"JG{}={};".format(self.acronym, velocity))
        self.begin_motion()

    @expose_method()
    def stop(self):
        self.logger.info(u"Stop motor channel movement")
        self.backlash_cleared_flag = True
        self.command(u"ST{};".format(self.acronym))
        self.set_value(FLAG_REFERENCING, False)



import asyncore
import logging
from collections import OrderedDict
from threading import Thread, Event, Lock, RLock
from time import sleep

import numpy

from kamzik3.constants import *
from kamzik3.snippets.snippetsTimer import PreciseCallbackTimer, CallbackTimer


class PortReadLoop(PreciseCallbackTimer):

    def __init__(self, interval=5):
        self.interval = interval
        self.logger = logging.getLogger(u"Console.PortReadLoop")
        PreciseCallbackTimer.__init__(self, interval, self.handle_interval_timeout)
        self.map = []
        self.stopped = Event()

    def run(self):
        self.logger.info(u"Starting port readout loop Thread")
        PreciseCallbackTimer.run(self)

    def add_device(self, device):
        if device not in self.map.copy():
            self.map.append(device)
            return True
        else:
            return False

    def remove_device(self, device):
        if device in self.map.copy():
            self.map.remove(device)
            return True
        else:
            return False

    def handle_interval_timeout(self):
        for device in self.map.copy():
            try:
                device.handle_read()
            except IOError:
                device.handle_response_error(u"Port read error")

    def stop(self):
        self.logger.info(u"Stopping port readout loop Thread")
        PreciseCallbackTimer.stop(self)


class DeviceConnectionPoller(CallbackTimer):

    def __init__(self, interval=2000):
        self.connecting_devices = []
        self.connected_devices = []
        self.interval = interval
        self.logger = logging.getLogger(u"Console.DeviceConnectionPoller")
        PreciseCallbackTimer.__init__(self, self.interval, self.check_devices_connection_timeout, with_correction=False)

    def run(self):
        self.logger.info(u"Starting devices connection poller Thread")
        CallbackTimer.run(self)

    def add_connecting_device(self, device):
        if device not in self.connecting_devices:
            self.connecting_devices.append(device)

    def remove_connecting_device(self, device):
        try:
            self.connecting_devices.remove(device)
        except ValueError:
            pass  # device was removed already

    def check_devices_connection_timeout(self):
        for device in self.connecting_devices.copy():
            if device.connecting_time >= device.connection_timeout:
                self.remove_connecting_device(device)
                device.handle_connection_error(u"Connection timeout")
            elif device.connected:
                self.remove_connecting_device(device)
                self.connected_devices.append(device)
            else:
                device.connecting_time += self.interval
        for device in self.connected_devices.copy():
            if not device.is_alive():
                self.connected_devices.remove(device)

    def stop(self):
        self.logger.info(u"Stopping devices connection poller Thread")
        CallbackTimer.stop(self)


class DevicePoller(PreciseCallbackTimer):

    def __init__(self, interval=5):
        self.interval = interval
        self.commands_buffer = {}
        self.poll_commands_buffer = {}
        self.poll_lock = RLock()
        self.command_lock = RLock()
        self.logger = logging.getLogger(u"Console.DevicePoller")
        self.polling_schedule = OrderedDict()
        self.schedule = []
        self.schedule_at = numpy.uint64(0)
        PreciseCallbackTimer.__init__(self, self.interval, self.timer_tick, with_correction=True)

    def run(self):
        self.logger.info(u"Starting devices poller Thread")
        PreciseCallbackTimer.run(self)

    def timer_tick(self):
        with self.command_lock:
            # Send pending commands
            for device, commands in self.commands_buffer.items():
                if commands and device.accepting_commands():
                    self.commands_buffer[device] = device.send_command(commands)

        with self.poll_lock:
            # Prepare polling commands
            for pollAt in self.schedule:
                if self.schedule_at % pollAt == 0:
                    for device, attributes in self.polling_schedule[pollAt].items():
                        if device.accepting_commands() and device in self.poll_commands_buffer:
                            self.poll_commands_buffer[device] += attributes

        with self.poll_lock:
            # Send polling commands
            for device, commands in self.poll_commands_buffer.items():
                if commands and device.accepting_commands():
                    self.poll_commands_buffer[device] = device.send_command(commands)
                    device.set_value(ATTR_BUFFERED_COMMANDS, len(self.poll_commands_buffer[device]))
        self.schedule_at += self.interval

    def set_timeout(self, timeout):
        PreciseCallbackTimer.set_timeout(self, timeout)
        self.interval = timeout
        self.schedule_at = 0

    def stop(self):
        self.logger.info(u"Stopping devices poller Thread")
        PreciseCallbackTimer.stop(self)

    def add(self, device, attribute, poll_at, callback=None, returning=True, force_add=False):
        with self.poll_lock:
            if poll_at not in self.polling_schedule:
                self.polling_schedule[poll_at] = OrderedDict()
            if device not in self.polling_schedule[poll_at]:
                self.polling_schedule[poll_at][device] = []
            if device not in self.poll_commands_buffer:
                self.poll_commands_buffer[device] = []

            polling_quadruplet = (attribute, None, callback, returning)
            if force_add or polling_quadruplet not in self.polling_schedule[poll_at][device]:
                self.polling_schedule[poll_at][device].append(polling_quadruplet)

            self.schedule = sorted(self.polling_schedule.keys())

    def remove(self, device, attribute, poll_at, callback=None, returning=True):
        with self.poll_lock:
            try:
                self.polling_schedule[poll_at][device].remove((attribute, None, callback, returning))
                if len(self.polling_schedule[poll_at][device]) == 0:
                    del self.polling_schedule[poll_at][device]
                    del self.poll_commands_buffer[device]
                if len(self.polling_schedule[poll_at]) == 0:
                    del self.polling_schedule[poll_at]

                self.schedule = sorted(self.polling_schedule.keys())
            except (ValueError, KeyError):
                pass

    def stop_polling(self, device):
        with self.poll_lock:
            if device in self.poll_commands_buffer:
                del self.poll_commands_buffer[device]
            for polledDevices in self.polling_schedule.values():
                try:
                    if device in polledDevices:
                        del polledDevices[device]
                except (ValueError, KeyError):
                    pass  # device is no longer within polled devices
        with self.command_lock:
            if device in self.commands_buffer:
                del self.commands_buffer[device]

    def prepare_command(self, device, command):
        with self.command_lock:
            try:
                self.commands_buffer[device].append(command)
            except KeyError:
                self.commands_buffer[device] = [command]

    def prepend_command(self, device, command):
        with self.command_lock:
            try:
                self.commands_buffer[device].insert(0, command)
            except KeyError:
                self.commands_buffer[device] = [command]


class DeviceAsyncoreLoop(Thread):
    dummy_device = None
    stopped = False

    def __init__(self):
        self.logger = logging.getLogger(u"Console.DeviceAsyncoreLoop")
        super(DeviceAsyncoreLoop, self).__init__()
        self.setDaemon(True)

    def run(self):
        self.logger.info(u"Starting devices asyncore loop Thread")
        self.stopped = False
        # We check if Loop is stopped
        while not self.stopped:
            # If not we start a loop. Loop is breaked when no devices are connected.
            asyncore.loop()
            # Sleep for a second before we tried to start asyncore loop again.
            sleep(1)

    def stop(self):
        self.logger.info(u"Stopping devices asyncore loop Thread")
        self.stopped = True
        asyncore.close_all()


control_asyncore_loop = DeviceAsyncoreLoop()
control_device_connection_poller = DeviceConnectionPoller()
control_device_poller = DevicePoller()
control_port_read_loop = PortReadLoop()



from asynchat import find_prefix_at_end

import serial

from kamzik3 import WriteException
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsControlLoops import control_port_read_loop


class DevicePort(Device):
    terminator = b"\r\n"
    ac_in_buffer_size = 4096
    ac_out_buffer_size = 4096
    port = None
    baud_rate = None
    parity = None
    stop_bits = None
    byte_size = None
    BAUDRATES = [50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
                 9600, 19200, 38400, 57600, 115200]

    def __init__(self, port, baud_rate=None, parity=None, stop_bits=None, byte_size=None, device_id=None, config=None):
        self.port = port
        self.baud_rate = baud_rate
        self.parity = parity
        self.stop_bits = stop_bits
        self.byte_size = byte_size
        self.port_read_loop = control_port_read_loop
        Device.__init__(self, device_id, config)
        self.ac_in_buffer = b''
        self.incoming = []
        self.buffer = []
        try:
            self.serial_port = serial.serial_for_url(self.port, baudrate=self.baud_rate, parity=self.parity,
                                                     stopbits=self.stop_bits, bytesize=self.byte_size, timeout=0)

            self.connect()
        except serial.SerialException:
            self.logger.exception(u"Serial port connection exception")
            self.connecting = True
            self.serial_port = None
            self.device_connection_poller.add_connecting_device(self)

    def connect(self):
        Device.connect(self)
        self.port_read_loop.add_device(self)

    def handshake(self):
        self.logger.info(u"Handshake initiated")
        return True

    def read_all(self):
        if self.serial_port.in_waiting > 0:
            if self.command_encoding is not None:
                return self.serial_port.read(self.serial_port.in_waiting).decode(self.command_encoding)
            else:
                return self.serial_port.read(self.serial_port.in_waiting)
        else:
            return None

    def handle_connect_event(self):
        if self.handshake():
            Device.handle_connect_event(self)

    def handle_connect(self):
        super(DevicePort, self).handle_connect()

    def push(self, data):
        if self.serial_port.write(data) != len(data):
            raise WriteException(u"Size of written data and written length is not equal.")

    def close_connection(self):
        Device.close_connection(self)

    def close(self):
        self.connected = False
        self.connecting = False
        self.closing = True
        self.port_read_loop.remove_device(self)

        if self.serial_port is not None:
            self.serial_port.close()

        self.closing = False
        self.closed = True
        self.set_status(STATUS_DISCONNECTED)
        # if self.response_error or self.connection_error:
        #     self.reconnect()

    def _collect_incoming_data(self, data):
        self.incoming.append(data)

    def _get_data(self):
        d = ''.join(self.incoming)
        del self.incoming[:]
        return d

    def set_terminator(self, term):
        """
        Set the input delimiter.
        :param term: str, int, None
        :return: 
        """
        self.terminator = term

    def get_terminator(self):
        return self.terminator

    # grab some more data from the socket,
    # throw it to the collector method,
    # check for the terminator,
    # if found, transition to the next state.

    def handle_read(self):

        if self.serial_port.in_waiting > 0:
            self.ac_in_buffer += self.serial_port.read(self.serial_port.in_waiting)
        else:
            return

        # Continue to search for self.terminator in self.ac_in_buffer,
        # while calling self.collect_incoming_data.  The while loop
        # is necessary because we might read several data+terminator
        # combos with a single recv(4096).

        while self.ac_in_buffer:
            lb = len(self.ac_in_buffer)
            terminator = self.get_terminator()
            if not terminator:
                # no terminator, collect it all
                self.collect_incoming_data(self.ac_in_buffer)
                self.ac_in_buffer = b''
            elif isinstance(terminator, int):
                # numeric terminator
                n = terminator
                if lb < n:
                    self.collect_incoming_data(self.ac_in_buffer)
                    self.ac_in_buffer = b''
                    self.terminator = self.terminator - lb
                else:
                    self.collect_incoming_data(self.ac_in_buffer[:n])
                    self.ac_in_buffer = self.ac_in_buffer[n:]
                    self.terminator = 0
                    self.found_terminator()
            else:
                # 3 cases:
                # 1) end of buffer matches terminator exactly:
                #    collect data, transition
                # 2) end of buffer matches some prefix:
                #    collect data to the prefix
                # 3) end of buffer does not match any prefix:
                #    collect data
                terminator_len = len(terminator)
                index = self.ac_in_buffer.find(terminator)
                if index != -1:
                    # we found the terminator
                    if index > 0:
                        # don't bother reporting the empty string
                        # (source of subtle bugs)
                        self.collect_incoming_data(self.ac_in_buffer[:index])
                    self.ac_in_buffer = self.ac_in_buffer[index + terminator_len:]
                    # This does the Right Thing if the terminator
                    # is changed here.
                    self.found_terminator()
                else:
                    # check for a prefix of the terminator
                    index = find_prefix_at_end(self.ac_in_buffer, terminator)
                    if index:
                        if index != lb:
                            # we found a prefix, collect up to the prefix
                            self.collect_incoming_data(self.ac_in_buffer[:-index])
                            self.ac_in_buffer = self.ac_in_buffer[-index:]
                        break
                    else:
                        # no prefix, collect it all
                        self.collect_incoming_data(self.ac_in_buffer)
                        self.ac_in_buffer = b''



import json
import pickle
import time
from threading import Thread, Event, RLock

import numpy as np
import zmq
from zmq.utils.monitor import recv_monitor_message

import kamzik3
from kamzik3 import DeviceClientError
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetDataAccess import get_from_dict, set_in_dict, is_equal
from kamzik3.snippets.snippetsJson import JsonKamzikEncoder, JsonKamzikHook

""" Example of yaml configuration
DeviceClient0: &DeviceClient0 !Device:kamzik3.devices.deviceClient.DeviceClient
    host: &Server 127.0.0.1
    port: &Port 60000
    device_id: DeviceIdOnServer
"""


class DeviceClient(Device):
    """
    : connect()
    : handle_connect_event()
            : handle_connect()
            : handle_configuration_event()
                : handle_configuration()
    """
    initial_status = STATUS_DISCONNECTED

    def __init__(self, host, port, device_id=None, config=None):
        self.subscription_thread = None
        Device.__init__(self, device_id, config)
        self.client_comm_lock = RLock()
        self.host = host
        self.port = port
        self.removing = False
        self.subscriber_port = None
        self.socket = None
        self.subscriber_socket = None
        self.monitor_socket = None
        self.monitor_thread = None
        self.callbacks = {}
        self.stopped = Event()
        self.subscribe_starts_with_list = []
        self.connect()

    def connect(self):
        """
        Here we want to connect first to ZMQ main server.
        We have to verify connection and do it in non blocking fashion.
        We just push devices in a pool and verify it in device_poller Thread.
        :return:
        """
        self.logger.info(u"Initiating connection to the server {}:{}".format(self.host, self.port))
        self.connecting = True
        self.connected = False
        self.set_status(STATUS_CONNECTING)
        # Get REQ type of socket
        self.socket = zmq.Context.instance().socket(zmq.REQ)
        # Set receive timeout to 3000 ms
        self.socket.setsockopt(zmq.RCVTIMEO, 3000)
        # This is VERY IMPORTANT not to set to anything but 0
        self.monitor_socket = self.socket.get_monitor_socket()
        self.monitor_thread = Thread(target=self.connect_monitor_thread)
        self.monitor_thread.start()
        try:
            self.socket.connect("tcp://{}:{}".format(self.host, self.port))
        except zmq.ZMQError:
            self.handle_connection_error(u"Fatal connection error")
            return

    def set_status(self, status):
        self._set((ATTR_STATUS, VALUE), status)

    def connect_monitor_thread(self):
        while not self.stopped.is_set():
            if self.monitor_socket.poll(timeout=500):
                message = recv_monitor_message(self.monitor_socket)
                if message["event"] == zmq.EVENT_CONNECTED:
                    try:
                        self.handle_connect_event()
                    except DeviceClientError:
                        # Server is ready, but device not yet configured
                        if not self.stopped.wait(5):
                            self.close()
                            self.reconnect()
                    return

    def handle_connect_event(self):
        """
        This method is called after we verified connection with main server.
        We continue with init method to subscribe to devices publisher.
        :return: None
        """
        if self.init():
            self.connected = True
            self.connecting = False
            self.handle_configuration_event()
            self.set_status(self.get_attribute([ATTR_STATUS, VALUE]))
            # Subscribe for all attribute changes
            token = self.get_token(TOKEN_ATTRIBUTE)
            self._subscribe(token)
            self.subscribe_starts_with_list = token
            # Check for attribute sharing and subscribe to them
            for attribute_id in self.attributes_sharing_map:
                self._subscribe(attribute_id)

    def _init_subscriber_thread(self, host, port):
        """
        We are connected to the main server.
        Now we want to connect to the devices publisher.
        :return: None
        """
        self.logger.info(u"Initializing {} client subscriber thread".format(self.device_id))
        # Get SUB type of socket
        self.subscriber_socket = zmq.Context.instance().socket(zmq.SUB)
        # Set receive timeout to 100 ms
        self.subscriber_socket.setsockopt(zmq.RCVTIMEO, 100)
        # Don't let socket linger after close
        # This is VERY IMPORTANT not to set to anything but 0
        self.subscriber_socket.setsockopt(zmq.LINGER, 0)
        self.subscriber_socket.connect("tcp://{}:{}".format(host, port))
        # Start separate Thread to read data from publisher
        self.subscription_thread = Thread(target=self._subscriber_thread)
        self.subscription_thread.start()

    def _subscriber_thread(self):
        """
        Continuously collect data from publisher.
        Interrupt every zmq.RCVTIMEO and check if client is connected.
        :return: None
        """
        # Check for published messages until client is connected
        while not self.stopped.isSet():
            if self.subscriber_socket.poll(timeout=100):
                # Try to read any message from publisher
                reply = self.subscriber_socket.recv_multipart()
                token, stype = reply[:2]
                if stype == MSG_JSON:
                    data = json.loads(reply[2].decode(), object_hook=JsonKamzikHook)
                elif stype == MSG_ARRAY:
                    dtype, shape = reply[2:4]
                    reply[4] = np.frombuffer(reply[4], dtype=dtype.decode())
                    data = np.reshape(reply[4], json.loads(shape.decode()))
                elif stype == MSG_PICKLE:
                    data = pickle.loads(reply[2])
                self.handle_readout(token.decode(), data)
            elif self.monitor_socket.poll(timeout=0):
                # Check if there is any message to read from monitor socket
                message = recv_monitor_message(self.monitor_socket)
                if message["event"] == zmq.EVENT_DISCONNECTED:
                    self.handle_response_error(u"Socket disconnected")
                    break

        # Client was disconnected for any reason, finish cleaning up and possibly reconnect
        self.handle_disconnect()

    def _subscribe(self, topic):
        """
        Subscribe for Topic.
        :param topic: string
        :return: None
        """
        self.subscriber_socket.setsockopt_string(zmq.SUBSCRIBE, topic)

    def _unsubscribe(self, topic):
        """
        Unsubscribe Topic.
        :param topic: string
        :return: None
        """
        self.subscriber_socket.setsockopt_string(zmq.UNSUBSCRIBE, topic)

    def set_callback(self, token, callback):
        """
        Set callback when token is received from publisher.
        If remove_after is set, remove callback from buffer and don't repeat it again.
        Callback has to be callable function.
        :param token: string
        :param callback: function(devices, data)
        :return: bool
        """
        assert callable(callback)

        if token not in self.callbacks:
            self.callbacks[token] = []

        callback_data = callback

        if callback_data not in self.callbacks[token]:
            self.callbacks[token].append(callback_data)
            self._subscribe(token)
            self.logger.debug(u"Set callback for token {}".format(token))
            return token
        else:
            self.logger.warning(u"Callback is already set")
            return False

    def handle_readout_callback(self, token, callback_data):
        """
        Handle callback for token.
        :param token: string
        :param callback_data: mixed
        :return: None
        """
        self._unsubscribe(token)
        while len(self.callbacks.get(token)) > 0:
            callback = self.callbacks.get(token).pop()
            callback(*callback_data)

        if len(self.callbacks[token]) == 0:
            del self.callbacks[token]

    def handle_readout(self, token, data):
        if token in self.callbacks:
            self.handle_readout_callback(token, data)
        else:
            attribute_parts = token.split(".")
            attribute_id = ".".join(attribute_parts[:-1])
            if attribute_parts[1] == TOKEN_ATTRIBUTE:
                if attribute_parts[-1] == TOKEN_ATTRIBUTE_REPLACE:
                    attribute_path = attribute_parts[2:-1]
                    attribute_path = self.attributes_sharing_map.get(attribute_id, attribute_path)

                    sync_tree = self.attributes
                    for part in attribute_path:
                        if part not in sync_tree:
                            sync_tree[part] = {}
                        sync_tree = sync_tree[part]
                    self._set(attribute_path, Attribute.from_dict(data))
                elif attribute_parts[-1] == TOKEN_ATTRIBUTE_DELETE:
                    attribute_path = attribute_parts[2:-1]
                    attribute_path = self.attributes_sharing_map.get(attribute_id, attribute_path)
                    sync_tree = self.attributes[attribute_path[:-1]]
                    del sync_tree[attribute_path[-1]]
                elif attribute_parts[-1] == TOKEN_ATTRIBUTE_GROUP_DELETE:
                    attribute_path = attribute_parts[2:-1]
                    del self.attributes[attribute_path[0]]
                else:
                    attribute_path = attribute_parts[2:]

                    if attribute_id in self.attributes_sharing_map:
                        for shared_attribute in self.attributes_sharing_map[attribute_id]:
                            self._set(shared_attribute + [attribute_parts[-1]], data)
                    if token.startswith(self.subscribe_starts_with_list):
                        self._set(attribute_path, data)

                # If publisher is defined, forward data to it
                if kamzik3.session.publisher is not None:
                    kamzik3.session.publisher.push_message(token, data)
            else:
                raise DeviceClientError(
                    "Unrecognized token part {}. Token value was: {}".format(attribute_parts[1], token))

    def init(self):
        """
        Initiate device connection.
        Get publisher port and devices attributes.
        :return:
        """
        try:
            with self.client_comm_lock:
                self.socket.send_multipart([INSTRUCTION_INIT, self.device_id.encode()])
                status, token, response = self.handle_server_response()
            if status == RESPONSE_OK:
                allow_proxy = self.config.get("allow_proxy", True)
                attributes, self.attributes_sharing_map, self.exposed_methods, self.qualified_name, device_proxy, device_publisher = json.loads(
                    response, object_hook=JsonKamzikHook)
                if device_proxy is not False and allow_proxy:
                    self.host, self.port = device_proxy
                    self.close()
                    self.reconnect()
                    return False
                self.initial_status = attributes[ATTR_STATUS][VALUE]
                self._sync_attributes(attributes, self.attributes)
                for method_name, attributes in self.exposed_methods[:]:
                    setattr(self, method_name,
                            lambda method_name=method_name, **kwargs: self.method(method_name, kwargs))
                self._init_subscriber_thread(*device_publisher)
                return True
            else:
                raise DeviceClientError(u"Failed to initialize device")

        except (zmq.Again, zmq.ZMQError):
            self.handle_response_error(u"Init error")
            self.reconnect()
            return False

    def _sync_attributes(self, synced_attributes, parent_attribute):
        """
        synchronize new attributes with already existing attributes.
        This is important to prevent creating new attribute objects.
        We can just create new attributes, but we would lost all callbacks and all references will become invalid.
        :param synced_attributes: dict
        :return:
        """
        # Walk through all items of dictionary
        for key, attribute in synced_attributes.items():
            self._sync_attribute(key, attribute, parent_attribute)

    def _sync_attribute(self, key, attribute, parent_attribute):
        """
        synchronize one attribute from new dict to existing self.attributes
        :param key: string
        :param attribute: dict
        :return:
        """
        if key in parent_attribute:
            if Attribute.is_attribute(attribute):
                if key == ATTR_STATUS and parent_attribute == self.attributes:
                    return
                """
                Key exists in self.attributes
                Prevent attribute changes when syncing with new values
                """
                for attr_key in attribute.keys():
                    if not is_equal(parent_attribute[key][attr_key], attribute[attr_key]):
                        parent_attribute[key][attr_key] = attribute[attr_key]
            else:
                # Attribute is only sub group, continue syncing deeper
                self._sync_attributes(attribute, parent_attribute[key])
        else:
            # Key des not exists in self.attributes
            if Attribute.is_attribute(attribute):
                # Create completely new Attribute under specified Key
                parent_attribute[key] = Attribute.from_dict(attribute)
            else:
                # Attribute is only sub group, continue syncing deeper
                parent_attribute[key] = Attribute.from_dict(attribute)

    def get_token(self, topic, device_id=None):
        return "{}.{}".format(self.device_id if device_id is None else device_id, topic)

    def set_attribute(self, attribute, value, callback=None):
        """
        Set attribute on devices server.
        :param attribute:
        :param value:
        :param callback:
        :return:
        """
        if not self.connected or self.closing:
            raise DeviceClientError(
                f"Failed to set remote device {self.device_id} {attribute} to {value}.\r\nDevice client is not connected")
        try:
            attribute_value = json.dumps((attribute, value), ensure_ascii=True)
            with self.client_comm_lock:
                self.socket.send_multipart([INSTRUCTION_SET, self.device_id.encode(), attribute_value.encode()])
                status, token, response = self.handle_server_response()
            if status == RESPONSE_OK:
                if callback is not None:
                    if token:
                        self.set_callback(self.get_token(token), callback)
                    else:
                        callback(attribute, value)
            else:
                raise DeviceClientError(
                    f"Failed to set remote device {self.device_id} {attribute} to {value} {self.get_attribute(attribute[:-1]).unit()}.\r\n{response}")

            return response

        except (zmq.Again, zmq.ZMQError):
            self.handle_response_error(u"Set error")
            return False

    def set_value(self, attribute, value, callback=None):
        """
        Set attribute value.
        :param attribute: list, dict, str
        :param value: mixed
        :return: None
        """
        attribute_list = Attribute.list_attribute(attribute)
        return self.set_attribute(attribute_list + [VALUE], value, callback)

    def set_raw_value(self, attribute, value):
        """
        Set attribute value obtained from device.
        Apply offset and factor before value is set.
        :param attribute: list, dict, str
        :param value: mixed
        :return: None
        """
        raise DeviceClientError(u"Set raw value is not implemented on Client."
                                u"Use set_value or set_attribute instead.")

    def _set(self, attribute, value, callback=None):
        """
        This sets Device attribute.
        Use this function when You want to set attribute by tuple or list key.
        Example: Device.set((ATTR_STATUS, VALUE), STATUS_IDLE)
        :param attribute: tuple, list, str
        :param value: mixed
        :return: None
        """
        current_value = get_from_dict(self.attributes, attribute)
        if not is_equal(value, current_value):
            set_in_dict(self.attributes, attribute, value)

    def command(self, command, callback=None):
        """
        Execute raw command directly on devices.
        :param command: string
        :param callback: function(devices, outputOfCommand)
        :return:
        """
        if not self.connected or self.closing:
            raise DeviceClientError(
                u"Remote command {0!r} execution failed. Client is not connected".format(command))

        try:
            with self.client_comm_lock:
                self.socket.send_multipart(
                    [INSTRUCTION_COMMAND, self.device_id.encode(), command.encode(), b"1" if callback else b"0"])
                status, token, response = self.handle_server_response()
            if status == RESPONSE_OK:
                if token and callback is not None:
                    self.set_callback(self.get_token(token), callback)
            else:
                raise DeviceClientError(u"Failed to execute remote command '{}'\n{}".format(command, response))
            return response
        except (zmq.Again, zmq.ZMQError) as e:
            self.handle_response_error(u"Command error: {}".format(e))
            return False

    def method(self, method, attributes=None):
        """
        Execute devices method with associated attributes.
        :param method: string
        :param attributes: json formatted string
        :return:
        """
        if not self.connected or self.closing:
            raise DeviceClientError(
                u"Method {} remote execution failed. Client is not connected".format(method))

        try:
            dumped_attributes = json.dumps(attributes, ensure_ascii=True, cls=JsonKamzikEncoder)
            with self.client_comm_lock:
                self.socket.send_multipart(
                    [INSTRUCTION_METHOD, self.device_id.encode(), str(method).encode(), dumped_attributes.encode()])
                status, token, response = self.handle_server_response()
            if status == RESPONSE_OK:
                return json.loads(response, object_hook=JsonKamzikHook)
            else:
                raise DeviceClientError(u"Failed to execute remote method '{}'\n{}".format(method, response))

        except (zmq.Again, zmq.ZMQError) as e:
            self.handle_response_error(u"Method error: {}".format(e))
            return False

    def handle_server_response(self):
        '''
        This is crucial moment here.
        If server takes more then RCVTIMEO time, then we get zmq error.
        In that case modify server response time or increase RCVTIMEO time.
        '''
        message = self.socket.recv_multipart()
        status, token, msg_type = message[:3]
        if msg_type == MSG_JSON:
            return status.decode(), token.decode(), message[3].decode()
        elif msg_type == MSG_FILE:
            return status.decode(), token.decode(), message[3].decode()

    def handle_configuration(self):
        start_at = time.time()
        self._config_attributes()
        self._config_commands()
        self.set_status(STATUS_CONFIGURED)
        self.set_status(self.initial_status)
        self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

    def handle_response_error(self, message=None):
        self.logger.warning(message)
        self.response_error = True
        self.close()

    def handle_connection_error(self, message=None):
        self.logger.error(message)
        self.connection_error = True
        self.disconnect()

    def handle_command_error(self, readout_command, readout_output):
        self.logger.error(
            u"Command error\nCommand: {}\nOutput: {}".format(readout_command, readout_output))

    def disconnect(self):
        self.close()

    def close(self):
        if self.logger is not None:
            self.logger.info(u"Closing client")
        self.closing = True
        if self.socket is not None:
            self.socket.disable_monitor()
        self.set_status(STATUS_DISCONNECTING)
        self.stopped.set()

    def handle_disconnect(self):
        if self.logger is not None:
            self.logger.info(u"Client closed")
        if not self.removing:
            self.closing = False
            self.connected = False
            self.set_status(STATUS_DISCONNECTED)
            if self.response_error or self.connection_error:
                self.reconnect()

    def remove(self):
        self.removing = True
        Device.remove(self)



import time
import numpy as np
import re

from kamzik3.snippets.snippetsDecorators import expose_method

from kamzik3.devices.deviceChannel import DeviceChannel
from serial import PARITY_NONE, STOPBITS_ONE, EIGHTBITS

import kamzik3
from kamzik3.constants import *
from kamzik3.devices.devicePort import DevicePort
from kamzik3.devices.deviceSocket import DeviceSocket


class DeviceMaxiGauge(DeviceSocket):
    terminator = b"\r\n"
    push_commands_max = 1
    measurement_statuses = {
        0: 'Measurement data okay',
        1: 'Underrange',
        2: 'Overrange',
        3: 'Sensor error',
        4: 'Sensor off',
        5: 'No sensor',
        6: 'Identification error'
    }
    ch_status = re.compile("^PR[1-9].$")
    channel_names = None

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.get_error_status()
        self.get_software_version()
        self.get_sensor_identification(callback=_finish_configuration)

    def _init_attributes(self):
        DeviceSocket._init_attributes(self)
        self.create_attribute(ATTR_MODEL, readonly=True)
        self.create_attribute(ATTR_CHANNELS, readonly=True)
        self.create_attribute(ATTR_CHANNELS_CONFIGURATION, readonly=True, default_type=TYPE_LIST)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = DeviceSocket.handle_readout(self, readout_buffer)
        if output == "\x15":
            self.handle_command_error(command, readout_buffer)
        elif command == 'PNR\r':
            self.set_value(ATTR_MODEL, output)
        elif command == 'ERR\r' and output != "0,0":
            err1, err2 = output.split(",")
            self.handle_command_error(command, f"Error code 1: {err1}, Error code 2: {err2}")
        elif command == 'TID\r':
            channels = output.split(",")
            self.set_value(ATTR_CHANNELS, len(channels))
            self.set_value(ATTR_CHANNELS_CONFIGURATION, channels)
        elif command == 'CID\r':
            self.channel_names = []
            for channel, name in enumerate(output.split(",")):
                self.notify((channel, ATTR_NAME), name)
                self.channel_names.append(name)
        elif self.ch_status.match(command):
            channel = int(command[2])
            status, pressure = output.split(",")
            self.notify((channel-1, ATTR_PRESSURE), float(pressure))
            self.notify((channel-1, ATTR_MEASUREMENT_STATUS), self.measurement_statuses[int(status)])

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(self.device_id, [command, readout_buffer], token)
            for observer in self._observers[:]:
                if isinstance(observer, DeviceMaxiGaugeChannel):
                    kamzik3.session.publisher.push_message(observer.device_id, [command, readout_buffer], token)

    def found_terminator(self):
        returning_flag = self.commands_buffer[0][0][3]
        if returning_flag and self.buffer[0] == "\x06":
            self.push(b"\x05\r")
            self.buffer = []
            return

        DeviceSocket.found_terminator(self)

    def get_software_version(self):
        self.command("PNR\r")

    def get_error_status(self):
        self.command("ERR\r")

    def get_sensor_identification(self, callback=None):
        self.command("TID\r", callback=callback)

    def start_polling(self):
        DeviceSocket.start_polling(self)
        self.poll_command("ERR\r", 1000)

class DeviceMaxiGaugePort(DevicePort, DeviceMaxiGauge):
    terminator = b"\r\n"

    def __init__(self, port, baud_rate=19200, parity=PARITY_NONE, stop_bits=STOPBITS_ONE, byte_size=EIGHTBITS,
                 device_id=None, config=None):
        DevicePort.__init__(self, port, baud_rate, parity, stop_bits, byte_size, device_id, config)

    def handshake(self):
        self.logger.info(u"Handshake initiated")
        # Ask for version
        self.push(b"PNR\r")
        # Give some time for devices to reply
        time.sleep(0.1)
        if self.serial_port.in_waiting != 0:
            if self.read_all().strip() == "\x06":
                # Request data
                self.push(b"\x05\r")
                # Give some time for devices to reply
                time.sleep(0.1)
                # Read reply
                self.read_all()
                return True
        return False

class DeviceMaxiGaugeChannel(DeviceChannel):

    def _init_attributes(self):
        DeviceChannel._init_attributes(self)
        self.create_attribute(ATTR_NAME, set_function=self.set_name)
        self.create_attribute(ATTR_PRESSURE, readonly=True, default_type=np.float, decimals=9, unit="mbar")
        self.create_attribute(ATTR_MEASUREMENT_STATUS, readonly=True)

    def handle_configuration(self):
        if self.configured:
            return

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.configured = True
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.connected = True
        self.get_sensor_name()
        _finish_configuration()

    def subject_update(self, key, value, subject):
        DeviceChannel.subject_update(self, key, value, subject)
        if self.connected and isinstance(key, tuple) and key[0] == self.channel:
            attribute = key[1]
            if attribute == ATTR_MEASUREMENT_STATUS:
                if value != "Sensor off":
                    self.set_status(STATUS_BUSY)
                else:
                    self.set_status(STATUS_IDLE)
            self.set_value(ATTR_LATENCY, self.device.get_value(ATTR_LATENCY))
            self.set_raw_value(attribute, value)

    def get_sensor_name(self):
        self.command("CID\r")

    def start_polling(self):
        DeviceChannel.start_polling(self)
        self.poll_command(f"PR{self.channel+1}\r", 300)

    def set_name(self, value):
        self.device.channel_names[self.channel] = value
        channel_names = ",".join(self.device.channel_names)
        self.command(f"CID,{channel_names}\r")

    @expose_method()
    def sensor_on(self):
        channels_count = self.device.get_value(ATTR_CHANNELS)
        channels_setup = ["0"] * channels_count
        channels_setup[self.channel] = "2"
        channels_setup = ",".join(channels_setup)
        self.command(f"SEN,{channels_setup}\r")

    @expose_method()
    def sensor_off(self):
        channels_count = self.device.get_value(ATTR_CHANNELS)
        channels_setup = ["0"] * channels_count
        channels_setup[self.channel] = "1"
        channels_setup = ",".join(channels_setup)
        self.command(f"SEN,{channels_setup}\r")

    @expose_method()
    def stop(self):
        self.sensor_off()


import os

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QMainWindow, QFileDialog

from kamzik3.constants import *
from kamzik3.gui.general.customOpenDialog import CustomOpenDialog
from kamzik3.gui.general.loadingOverlayWidget import LoadingOverlayWidget
from kamzik3.gui.templates.logViewWindowTemplate import Ui_MainWindow
from kamzik3.snippets.snippetsWidgets import show_error_message


class LogViewWindow(QMainWindow, Ui_MainWindow):
    sig_task_update = pyqtSignal(str, str, object, object)
    sig_task_progress_update = pyqtSignal(float)

    def __init__(self, view_widget, cache_directory="./.cache/", virtual_directory_device=None):
        self.virtual_directory_device = virtual_directory_device
        self.cache_directory = cache_directory
        QMainWindow.__init__(self)
        self.setupUi(self)
        self.setWindowIcon(QIcon(":/icons/icons/log_icon_client.png"))
        self.setCentralWidget(view_widget)
        self.sig_task_update.connect(self.slot_task_update)
        self.loading_overlay = LoadingOverlayWidget(self)
        self.sig_task_progress_update.connect(self.loading_overlay.slot_set_progress)

    def load_file(self):
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog

        if self.virtual_directory_device is None:
            extension_filter = "Logfile (*.log *.log.*)"
            file_path, _ = QFileDialog.getOpenFileName(self, 'Choose recipe to load', ".", filter=extension_filter,
                                                       initialFilter=extension_filter, options=options)
            if file_path != "":
                self.slot_open_file(file_path, file_path)
        else:
            if self.virtual_directory_device.get_value(ATTR_STATUS) in READY_DEVICE_STATUSES:
                dialog = CustomOpenDialog(self.virtual_directory_device, parent=self)
                if dialog.exec():
                    requested_files, metadata = self.virtual_directory_device.prepare_files(input_files=dialog.selected_files)
                    task_id = self.virtual_directory_device.sync_files(requested_files)

                    def callback(value):
                        self.sig_task_update.emit(value, task_id, dialog.selected_files, metadata)

                    def cancel():
                        self.virtual_directory_device.stop_task(task_id)

                    self.loading_overlay.set_over(self, cancel_function=cancel, message="Syncing...")
                    self.virtual_directory_device.attach_attribute_callback([task_id, ATTR_STATUS],
                                                                            callback, key_filter=VALUE)
                    self.virtual_directory_device.attach_attribute_callback([task_id, ATTR_PROGRESS],
                                                                            self.sig_task_progress_update.emit,
                                                                            key_filter=VALUE)
            else:
                show_error_message("Virtual directory device is not ready.", parent=self)

    def slot_task_update(self, status, task_id, original_file_names, metadata):
        if status == STATUS_BUSY:
            pass
        elif status == STATUS_IDLE:
            aborted = self.virtual_directory_device.get_value([task_id, ATTR_ABORTED])
            self.loading_overlay.hide()
            if not aborted:
                synced_files = self.virtual_directory_device.get_value([task_id, ATTR_SYNCED_FILES])
                self.centralWidget().input_windowed_plots.setChecked(metadata.get("windowed", True))
                self.centralWidget().input_view_scale_plots.setChecked(metadata.get("scaled", True))
                for index, value in enumerate(synced_files.items()):
                    _, local_path = value
                    self.centralWidget().open(local_path, append=index != 0, refresh=index == 0,
                                              filter=metadata.get("filter", None))
                self.centralWidget().slot_select_rows(metadata.get("select", None))
                self.setWindowTitle(os.path.join(*original_file_names[0]))

    def slot_open_file(self, local_path, file_path):
        self.centralWidget().open(local_path, append=False, refresh=True)
        self.setWindowTitle(file_path)

    def resizeEvent(self, QResizeEvent):
        self.loading_overlay.resizeEvent(QResizeEvent)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'deviceRheodyneTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.10.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(855, 98)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Form.sizePolicy().hasHeightForWidth())
        Form.setSizePolicy(sizePolicy)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 0, 2, 1, 1)
        self.layout_switch_list = QtWidgets.QGridLayout()
        self.layout_switch_list.setObjectName("layout_switch_list")
        self.gridLayout.addLayout(self.layout_switch_list, 1, 0, 1, 4)
        self.label_device_name = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_device_name.sizePolicy().hasHeightForWidth())
        self.label_device_name.setSizePolicy(sizePolicy)
        self.label_device_name.setMinimumSize(QtCore.QSize(150, 0))
        self.label_device_name.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setFamily("Sans Serif")
        font.setPointSize(9)
        font.setBold(True)
        font.setWeight(75)
        self.label_device_name.setFont(font)
        self.label_device_name.setStyleSheet("")
        self.label_device_name.setScaledContents(False)
        self.label_device_name.setObjectName("label_device_name")
        self.gridLayout.addWidget(self.label_device_name, 0, 0, 1, 1)
        self.button_debug = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_debug.sizePolicy().hasHeightForWidth())
        self.button_debug.setSizePolicy(sizePolicy)
        self.button_debug.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/settings.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_debug.setIcon(icon)
        self.button_debug.setIconSize(QtCore.QSize(19, 19))
        self.button_debug.setObjectName("button_debug")
        self.gridLayout.addWidget(self.button_debug, 0, 3, 1, 1)
        self.label_status = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_status.sizePolicy().hasHeightForWidth())
        self.label_status.setSizePolicy(sizePolicy)
        self.label_status.setMinimumSize(QtCore.QSize(100, 0))
        self.label_status.setMaximumSize(QtCore.QSize(16777215, 16777215))
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        font.setKerning(True)
        self.label_status.setFont(font)
        self.label_status.setStyleSheet("QLabel {background:#68d532}")
        self.label_status.setAlignment(QtCore.Qt.AlignCenter)
        self.label_status.setObjectName("label_status")
        self.gridLayout.addWidget(self.label_status, 0, 1, 1, 1)
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setContentsMargins(-1, -1, 6, 0)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.label_macro_list = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_macro_list.sizePolicy().hasHeightForWidth())
        self.label_macro_list.setSizePolicy(sizePolicy)
        self.label_macro_list.setStyleSheet("font-weight:bold;background:#f6b442;padding:5")
        self.label_macro_list.setObjectName("label_macro_list")
        self.gridLayout_2.addWidget(self.label_macro_list, 0, 0, 1, 1)
        self.layout_macro_grid = QtWidgets.QGridLayout()
        self.layout_macro_grid.setSizeConstraint(QtWidgets.QLayout.SetMinimumSize)
        self.layout_macro_grid.setObjectName("layout_macro_grid")
        self.gridLayout_2.addLayout(self.layout_macro_grid, 1, 0, 1, 1)
        self.gridLayout.addLayout(self.gridLayout_2, 2, 0, 1, 5)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem1, 3, 0, 1, 1)

        self.retranslateUi(Form)
        self.button_debug.clicked.connect(Form.show_debug_widget)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label_device_name.setText(_translate("Form", "ID"))
        self.button_debug.setToolTip(_translate("Form", "Settings"))
        self.label_status.setText(_translate("Form", "STATUS"))
        self.label_macro_list.setText(_translate("Form", "Macros"))

from kamzik3 import resource_kamzik3_rc


import random
import time
from threading import Thread
from time import sleep

import numpy as np

from kamzik3.constants import *
from kamzik3.devices.dummy.deviceDummy import DeviceDummy

ATTR_CHAMELEON = u"Chameleon"
ATTR_RANDOM_VALUE = u"Random value"
ATTR_CHAMELEON_UNIT = u"Chameleon unit"
ATTR_GROUP_ATTRIBUTE = u"Group attribute"
ATTR_SHARED_ATTRIBUTE = u"Shared attribute"
ATTR_SHARED_VALUE = u"Shared value"
GROUP_TEST_GROUP = u"Test group"


class DummyFeatureTestDevice(DeviceDummy):

    def _init_attributes(self):
        DeviceDummy._init_attributes(self)
        a = self.create_attribute(ATTR_CHAMELEON, default_value=0, default_type=np.float, decimals=3, unit="")
        self.create_attribute(ATTR_CHAMELEON_UNIT, GROUP_TEST_GROUP, default_value="",
                              set_function=self.set_chameleon_unit, save_change=True)
        r = self.create_attribute(ATTR_RANDOM_VALUE, default_value=0, readonly=True, default_type=np.float, decimals=3,
                                  unit="", min_value=50, max_value=950)
        self.create_attribute(ATTR_GROUP_ATTRIBUTE, GROUP_TEST_GROUP, default_value=0, default_type=np.float,
                              decimals=3, unit="")
        self.add_attribute(ATTR_SHARED_ATTRIBUTE, r)
        self.add_attribute(ATTR_SHARED_VALUE, a)

    def handle_configuration(self):
        start_at = time.time()
        self._config_commands()
        self._config_attributes()
        Thread(target=self.random_value_generator).start()
        self.set_status(STATUS_CONFIGURED)
        self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

    def set_chameleon_unit(self, value):
        self.set_attribute([ATTR_CHAMELEON, UNIT], value)
        self.set_attribute([ATTR_RANDOM_VALUE, UNIT], value)
        self.set_attribute([GROUP_TEST_GROUP, ATTR_GROUP_ATTRIBUTE, UNIT], value)

    def random_value_generator(self):
        while self.connected:
            value = random.uniform(0, 1e3)
            self.set_attribute([ATTR_RANDOM_VALUE, VALUE], value)
            sleep(0.2)



import numpy as np
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtWidgets import QWidget

from kamzik3 import units
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.gui.attributeDisplayWidget import AttributeDisplayWidget
from kamzik3.snippets.snippetsYaml import YamlSerializable

class MacroWidget(QWidget, YamlSerializable):

    user_mode = MODE_SIMPLE

    def __init__(self, title=None, config=None, user_mode=MODE_SIMPLE, parent=None):
        self.title = title
        self.user_mode = user_mode
        self.macro_inputs = {}
        self.macro_widgets = {}
        self.config = config
        QWidget.__init__(self, parent)

    def yaml_mapping(self):
        mapping = super().yaml_mapping()
        del mapping["parent"]
        return mapping

    def set_title(self, title=None):
        if title is None:
            title = ""
        self.title = "{} scan".format(title)
        self.label_axis_settings.setText(self.title)

    def _show_method_attribute_widget(self, attribute_title, attribute, row):
        if attribute is None:
            attribute = Attribute("")
        if self.config.get(attribute_title, False) and self.config.get(attribute_title, None) is not None:
            attribute[VALUE] = self.config[attribute_title]
        attribute_widget = AttributeDisplayWidget(attribute_title, attribute, config={"label_width": 110})
        self.layout().addWidget(attribute_widget.label_widget, row, 0)
        self.layout().addWidget(attribute_widget.input_widget, row, 1)
        self.layout().addWidget(attribute_widget.unit_widget, row, 2)
        self.macro_inputs[attribute_title] = attribute_widget
        return attribute_widget

    def _show_attribute_widget(self, row):
        attribute = self.attribute.attribute_copy()
        attribute_widget = AttributeDisplayWidget(u"Value", attribute, config={"label_width": 110})
        self.layout().addWidget(attribute_widget.label_widget, row, 0)
        self.layout().addWidget(attribute_widget.input_widget, row, 1)
        self.layout().addWidget(attribute_widget.unit_widget, row, 2)
        self.macro_inputs["value"] = attribute_widget
        return attribute_widget

    def _show_start_point_widget(self, row, default_unit=None):
        attribute = self.attribute.attribute_copy()
        if default_unit is not None and attribute.value() is not None:
            default_value = units.Quantity(self.config.get("end_point", attribute.value())).to(default_unit)
        else:
            default_value = self.config.get("end_point", attribute.value())
        attribute[VALUE] = default_value
        start_point_widget = AttributeDisplayWidget(u"Start point", attribute,
                                                    config={"label_width": 110})
        if default_unit is not None:
            start_point_widget.set_unit(default_unit)
        self.layout().addWidget(start_point_widget.label_widget, row, 0)
        self.layout().addWidget(start_point_widget.input_widget, row, 1)
        self.layout().addWidget(start_point_widget.unit_widget, row, 2)
        self.macro_inputs["start_point"] = start_point_widget
        start_point_widget.input_widget.valueChanged.connect(self.signal_start_end_changed)
        return start_point_widget

    def _show_end_point_widget(self, row, default_unit=None):
        attribute = self.attribute.attribute_copy()
        if default_unit is not None and attribute.value() is not None:
            default_value = units.Quantity(self.config.get("end_point", attribute.value())).to(default_unit)
        else:
            default_value = self.config.get("end_point", attribute.value())
        attribute[VALUE] = default_value
        end_point_widget = AttributeDisplayWidget(u"End point", attribute, config={"label_width": 110})
        if default_unit is not None:
            end_point_widget.set_unit(default_unit)
        self.end_point_lock_button.toggled.connect(end_point_widget.input_widget.setDisabled)
        self.end_point_set_widget.layout().addWidget(end_point_widget.input_widget)
        self.layout().addWidget(end_point_widget.label_widget, row, 0)
        self.layout().addWidget(self.end_point_set_widget, row, 1)
        self.layout().addWidget(end_point_widget.unit_widget, row, 2)
        self.macro_inputs["end_point"] = end_point_widget
        end_point_widget.input_widget.valueChanged.connect(self.signal_start_end_changed)
        end_point_widget.input_widget.editingFinished.connect(self.signal_start_end_check_limits)
        return end_point_widget

    def _show_field_size_widget(self, row, default_unit=None):
        field_size_attribute = self.attribute.attribute_copy()
        default_value = 0
        if default_unit is not None:
            default_value = units.Quantity(default_value, default_unit)
        field_size_attribute[VALUE] = default_value
        field_size_attribute[MAX] = abs(field_size_attribute[MAX] - field_size_attribute[MIN])
        field_size_attribute[MIN] = 0
        field_size_widget = AttributeDisplayWidget(u"Field size", field_size_attribute,
                                                   config={"label_width": 110})
        if default_unit is not None:
            field_size_widget.set_unit(default_unit)
        self.field_size_lock_button.toggled.connect(field_size_widget.input_widget.setDisabled)
        self.field_size_set_widget.layout().addWidget(field_size_widget.input_widget)
        self.macro_inputs["field_size"] = field_size_widget
        field_size_widget.input_widget.setDisabled(self.field_size_lock_button.isChecked())
        self.layout().addWidget(field_size_widget.label_widget, row, 0)
        self.layout().addWidget(self.field_size_set_widget, row, 1)
        self.layout().addWidget(field_size_widget.unit_widget, row, 2)
        field_size_widget.input_widget.valueChanged.connect(self.signal_field_size_changed)
        field_size_widget.input_widget.editingFinished.connect(self.signal_field_size_check_limits)
        return field_size_widget

    def _show_step_size_widget(self, row, default_unit=None):
        step_size_attribute = self.attribute.attribute_copy()
        default_value = 0
        if default_unit is not None:
            default_value = units.Quantity(default_value, default_unit)
        step_size_attribute[VALUE] = default_value
        if self.radio_absolute_scan.isChecked():
            distance = abs(step_size_attribute[MAX] - step_size_attribute[MIN])
            step_size_attribute[MIN] = -distance
            step_size_attribute[MAX] = distance
        step_size_widget = AttributeDisplayWidget(u"Step size", step_size_attribute, config={"label_width": 110})
        if default_unit is not None:
            step_size_widget.set_unit(default_unit)
        self.step_size_set_widget.layout().addWidget(step_size_widget.input_widget)
        self.step_size_lock_button.toggled.connect(step_size_widget.input_widget.setDisabled)
        step_size_widget.input_widget.setDisabled(self.step_size_lock_button.isChecked())
        self.layout().addWidget(self.step_size_set_widget, row, 1)

        if self.radio_relative_scan.isChecked():
            step_size_widget.input_widget.valueChanged.connect(self.signal_step_size_changed)
            step_size_widget.input_widget.editingFinished.connect(self.signal_step_size_check_limits)
        else:
            step_size_widget.input_widget.valueChanged.connect(self.signal_absolute_step_size_changed)
            step_size_widget.input_widget.editingFinished.connect(self.signal_absolute_step_size_check_limits)
        self.layout().addWidget(step_size_widget.label_widget, row, 0)
        self.layout().addWidget(step_size_widget.unit_widget, row, 2)
        self.macro_inputs["step_size"] = step_size_widget
        return step_size_widget

    def _show_total_steps_widget(self, row):
        default_value = self.config.get("total_steps", 10)
        total_steps_widget = AttributeDisplayWidget("Total steps",
                                                    Attribute(default_value=default_value, min_value=0,
                                                              default_type=np.uint64),
                                                    config={"label_width": 110})
        self.total_steps_set_widget.layout().addWidget(total_steps_widget.input_widget)
        self.total_steps_lock_button.toggled.connect(total_steps_widget.input_widget.setDisabled)
        total_steps_widget.input_widget.setDisabled(self.total_steps_lock_button.isChecked())
        self.layout().addWidget(self.total_steps_set_widget, row, 1)

        if self.radio_relative_scan.isChecked():
            total_steps_widget.input_widget.valueChanged.connect(self.signal_total_steps_changed)
            total_steps_widget.input_widget.editingFinished.connect(self.signal_total_steps_check_limits)
        else:
            total_steps_widget.input_widget.valueChanged.connect(self.signal_absolute_total_steps_changed)
            total_steps_widget.input_widget.editingFinished.connect(self.signal_absolute_total_steps_check_limits)
        self.step_size_lock_button.setChecked(False)
        self.layout().addWidget(total_steps_widget.label_widget, row, 0)
        self.macro_inputs["total_steps"] = total_steps_widget
        self.layout().addWidget(total_steps_widget.unit_widget, row, 2)
        return total_steps_widget

    def _show_negative_tolerance_widget(self, row, default_unit=None):
        negative_tolerance_attribute = self.attribute.attribute_copy()
        negative_tolerance_attribute[VALUE] = self.attribute[TOLERANCE][0]
        negative_tolerance_widget = AttributeDisplayWidget("Negative tolerance", negative_tolerance_attribute,
                                                           config={"label_width": 110})
        negative_tolerance_widget.input_widget.setPrefix("- ")
        if default_unit is not None:
            negative_tolerance_widget.set_unit(default_unit)
        self.layout().addWidget(negative_tolerance_widget.label_widget, row, 0)
        self.layout().addWidget(negative_tolerance_widget.input_widget, row, 1)
        self.layout().addWidget(negative_tolerance_widget.unit_widget, row, 2)
        self.macro_inputs["negative_tolerance"] = negative_tolerance_widget
        return negative_tolerance_widget

    def _show_positive_tolerance_widget(self, row, default_unit=None):
        positive_tolerance_attribute = self.attribute.attribute_copy()
        positive_tolerance_attribute[VALUE] = self.attribute[TOLERANCE][1]
        positive_tolerance_widget = AttributeDisplayWidget("Positive tolerance", positive_tolerance_attribute,
                                                           config={"label_width": 110})
        positive_tolerance_widget.input_widget.setPrefix("+ ")
        if default_unit is not None:
            positive_tolerance_widget.set_unit(default_unit)
        self.layout().addWidget(positive_tolerance_widget.label_widget, row, 0)
        self.layout().addWidget(positive_tolerance_widget.input_widget, row, 1)
        self.layout().addWidget(positive_tolerance_widget.unit_widget, row, 2)
        self.macro_inputs["positive_tolerance"] = positive_tolerance_widget
        return positive_tolerance_widget

    def _show_random_deviation_widget(self, row):
        default_value = units.Quantity(self.config.get("Random deviation", "0 percent")).to("percent")
        random_deviation_widget = AttributeDisplayWidget("Random deviation",
                                                         Attribute(default_value=default_value, decimals=2, default_type=np.float16,
                                                                   min_value=0,
                                                                   max_value=100, unit=u"%"),
                                                         config={"static_unit": True, "label_width": 110})
        self.layout().addWidget(random_deviation_widget.label_widget, row, 0)
        self.layout().addWidget(random_deviation_widget.input_widget, row, 1)
        self.layout().addWidget(random_deviation_widget.unit_widget, row, 2)
        self.macro_inputs["random_deviation"] = random_deviation_widget
        return random_deviation_widget

    def _show_wait_after_widget(self, row):
        default_value = units.Quantity(self.config.get("wait_after", "0s")).to("s")
        wait_after_widget = AttributeDisplayWidget("Wait after",
                                                   Attribute(default_value=default_value, decimals=0, default_type=np.float64,
                                                             min_value=0,
                                                             unit=u"ms",
                                                             description="Set waiting time after each step."),
                                                   config={"static_unit": False, "label_width": 110})
        self.layout().addWidget(wait_after_widget.label_widget, row, 0)
        self.layout().addWidget(wait_after_widget.input_widget, row, 1)
        self.layout().addWidget(wait_after_widget.unit_widget, row, 2)
        self.macro_inputs["wait_after"] = wait_after_widget
        return wait_after_widget

    def _show_repeat_widget(self, row):
        repeat_widget = AttributeDisplayWidget("Repeat", Attribute(default_value=0, default_type=np.uint64,
                                                                   description="Repeat step after first execution. Set number of repeats.\r\nEach step is executed at least ones"),
                                               config={"label_width": 110})
        self.layout().addWidget(repeat_widget.label_widget, row, 0)
        self.layout().addWidget(repeat_widget.input_widget, row, 1)
        self.layout().addWidget(repeat_widget.unit_widget, row, 2)
        self.macro_inputs["repeat"] = repeat_widget
        return repeat_widget

    def _show_retry_widget(self, row):
        default_value = self.config.get("retry", 0)
        retry_widget = AttributeDisplayWidget("Retry step",
                                              Attribute(default_value=int(default_value), default_type=np.uint32,
                                                        min_value=0,
                                                        description="In case of error, retry_count step again.\r\nSet maximum number of retries before throwing an error."),
                                              config={"static_unit": True, "label_width": 110})
        self.layout().addWidget(retry_widget.label_widget, row, 0)
        self.layout().addWidget(retry_widget.input_widget, row, 1)
        self.layout().addWidget(retry_widget.unit_widget, row, 2)
        self.macro_inputs["retry_count"] = retry_widget
        return retry_widget

    def _show_busy_timeout(self, row):
        default_value = units.Quantity(self.config.get("busy_timeout", "1000ms")).to("ms")
        busy_timeout_widget = AttributeDisplayWidget("Busy timeout",
                                                     Attribute(default_value=default_value, decimals=0,
                                                               default_type=np.float64,
                                                               min_value=0,
                                                               unit=u"ms",
                                                               description="Maximum waiting time for device to change from IDLE to BUSY status. Otherwise consider step as DONE."),
                                                     config={"static_unit": False, "label_width": 110})
        self.layout().addWidget(busy_timeout_widget.label_widget, row, 0)
        self.layout().addWidget(busy_timeout_widget.input_widget, row, 1)
        self.layout().addWidget(busy_timeout_widget.unit_widget, row, 2)
        self.macro_inputs["busy_timeout"] = busy_timeout_widget
        return busy_timeout_widget

    def _show_timeout(self, row):
        default_value = units.Quantity(self.config.get("timeout", "0s")).to("s")
        timeout_widget = AttributeDisplayWidget("Step timeout",
                                                Attribute(default_value=default_value, decimals=0,
                                                          default_type=np.float64,
                                                          min_value=0,
                                                          unit=u"s",
                                                          description="Maximum waiting time for step to finish."),
                                                config={"static_unit": False, "label_width": 110})
        self.layout().addWidget(timeout_widget.label_widget, row, 0)
        self.layout().addWidget(timeout_widget.input_widget, row, 1)
        self.layout().addWidget(timeout_widget.unit_widget, row, 2)
        self.macro_inputs["timeout"] = timeout_widget
        return timeout_widget

    def _show_on_warning_widget(self, row):
        default_value = self.config.get("on_warning", "Abort macro")
        on_warning_widget = AttributeDisplayWidget("On warning",
                                              Attribute(default_value=default_value, default_type=["Abort macro", "Skip"],
                                                        description="In case of warning, do specific action."),
                                              config={"static_unit": True, "label_width": 110})
        self.layout().addWidget(on_warning_widget.label_widget, row, 0)
        self.layout().addWidget(on_warning_widget.input_widget, row, 1)
        self.layout().addWidget(on_warning_widget.unit_widget, row, 2)
        self.macro_inputs["on_warning"] = on_warning_widget
        return on_warning_widget

    def get_input_values(self):
        output = {}
        for key, input in self.macro_inputs.items():
            output[key] = input.get_attribute_value()
        return output

    def get_link(self):
        raise NotImplementedError()

    @pyqtSlot()
    def set_simple_mode(self):
        self.user_mode = MODE_SIMPLE

    @pyqtSlot()
    def set_extended_mode(self):
        self.user_mode = MODE_EXTENDED

    @pyqtSlot()
    def set_all_mode(self):
        self.user_mode = MODE_ALL


import logging

from PyQt5.QtCore import pyqtSignal, pyqtSlot
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QWidget

import kamzik3
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.devices.observer import Observer
from kamzik3.snippets.snippetsWidgets import show_error_message
from kamzik3.snippets.snippetsYaml import YamlSerializable


class DeviceWidget(Observer, QWidget, YamlSerializable):
    sigHandleConfiguration = pyqtSignal()
    sigEnable = pyqtSignal("bool")
    sigStatus = pyqtSignal("PyQt_PyObject")
    sigErrorCheck = pyqtSignal("PyQt_PyObject", "PyQt_PyObject")
    sigShowModel = pyqtSignal("QIcon")
    sigHideModel = pyqtSignal("QIcon")
    device = None
    status_colors = {
        STATUS_DISCONNECTED: "silver",
        STATUS_CONNECTING: "#d532cc",
        STATUS_CONFIGURING: "#d532cc",
        STATUS_CONFIGURED: "#68d532",
        STATUS_IDLE: "#68d532",
        STATUS_ERROR: "#d53232",
        STATUS_BUSY: "#3290d5",
        STATUS_DISCONNECTING: "silver"
    }
    model_image = None
    config = None
    configured = False

    def __init__(self, device, model_image=None, config=None, parent=None):
        QWidget.__init__(self)
        Observer.__init__(parent)
        if isinstance(device, Device):
            self.device_id = device.device_id
            self.device = device
        else:
            self.device_id = device
            self.device = kamzik3.session.get_device(device)
        if model_image is not None:
            self.model_image = QIcon(model_image)
        self.config = config
        if self.config is None:
            self.config = {}
        self.logger = logging.getLogger("Gui.Device.{}".format(self.device_id))
        self.setupUi(self)
        self.init_signals()
        self.attach_to_subject(self.device)

    def init_signals(self):
        self.sigHandleConfiguration.connect(self.slot_handle_configuration)
        self.sigEnable.connect(self.slot_set_enabled)
        self.sigStatus.connect(self.slot_set_status)
        self.sigErrorCheck.connect(self.slot_check_error)

    def yaml_mapping(self):
        mapping = super().yaml_mapping()
        del mapping["parent"]
        return mapping

    @pyqtSlot("PyQt_PyObject", "PyQt_PyObject")
    def slot_check_error(self, attribute, output):
        if attribute == RESPONSE_ERROR:
            self.logger.error(output)
            show_error_message(output, parent=self)

    @pyqtSlot()
    def slot_handle_configuration(self):
        """
        Handle widget configuration.
        If overloading make sure to set configured flag.
        :return:
        """
        self.configured = True

    @pyqtSlot("bool")
    def slot_set_enabled(self, value):
        raise NotImplementedError(u"Not implemented")

    @pyqtSlot("PyQt_PyObject")
    def slot_set_status(self, value):
        """
        This slot is always called when device status has changed.
        Recall this method when overloading.
        :param value:
        :return:
        """
        if not self.configured and value in READY_DEVICE_STATUSES:
            self.slot_handle_configuration()

    def on_status_changed(self, key, value):
        if key == VALUE:
            self.sigStatus.emit(value)

    def on_enabled_changed(self, key, value):
        if key == VALUE:
            self.sigEnable.emit(value)

    def subject_update(self, key, value, subject):
        """
        This method is called by observing device.
        Useual key is ATTR_STATUS.
        :param key: mixed
        :param value: mixed
        :param subject: Subject
        :return:
        """
        if key == ATTR_STATUS:
            self.on_status_changed(VALUE, value)

    def get_status_color(self, status=None):
        """
        Get color for each STATUS.
        Overload status_colors if You want to use different status colors.
        :param status: string
        :return: status color
        """
        if status is None:
            status = self.device.get_value(ATTR_STATUS)

        return self.status_colors.get(status, "blue")

    def show_setpoints_history(self):
        print("test", self.input_value.setpoints_history)

    def close(self):
        self.detach_subject(self.device)
        self.device.detach_observer(self)
        self.device = None
        self.model_image = None
        self.sigHandleConfiguration.disconnect()
        self.sigEnable.disconnect()
        self.sigStatus.disconnect()
        self.sigErrorCheck.disconnect()
        self.setParent(None)
        QWidget.close(self)



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QCheckBox, QSizePolicy

from kamzik3.constants import *
from kamzik3.gui.attributeWidgets.attributeWidget import AttributeWidget


class AttributeBoolWidget(AttributeWidget):
    attribute_type_cast = bool

    def _set_input_widget(self):
        self.input_widget = QCheckBox()
        self.input_widget.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.input_widget.setStyleSheet("QCheckBox:checked {color: green} QCheckBox:unchecked {color: red}")
        self.input_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.input_widget.clicked.connect(self.clicked)
        if self.attribute[READONLY]:
            self.input_widget.setAttribute(Qt.WA_TransparentForMouseEvents)
            self.input_widget.setFocusPolicy(Qt.NoFocus)

    def clicked(self, flag):
        self.input_widget.setText("On" if flag else "Off")

    def set_value(self, value):
        if value is None:
            return
        self.input_widget.blockSignals(True)
        self.input_widget.setText("On" if value else "Off")
        self.input_widget.setChecked(value)
        self.input_widget.blockSignals(False)

    def get_attribute_value(self):
        return self.input_widget.isChecked()

    def get_widget_value(self):
        return self.input_widget.isChecked()



import time

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.device import Device


class DeviceDummy(Device):

    def __init__(self, device_id=None, config=None):
        Device.__init__(self, device_id, config)
        self.connect()

    def connect(self, *args):
        """
        Call only this function to connect devices to port / socket / library / ...
        :param args: connect attributes
        """
        self.connecting = True
        self.connected = True
        self.handle_configuration_event()
        self.connecting = False
        self.set_status(STATUS_IDLE)

    def handle_configuration(self):
        start_at = time.time()
        self._config_attributes()
        self.set_status(STATUS_CONFIGURED)
        self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

    def command(self, command, callback=None, with_token=False, returning=True):
        raise DeviceError("Dummy device does not accept any commands.")

    def handshake(self):
        return True

    def start_polling(self):
        return

    def close(self):
        Device.close(self)


import logging
import statistics
import sys
import traceback
from logging import Handler
from logging.handlers import TimedRotatingFileHandler
from time import time, perf_counter_ns, sleep

import numpy

import kamzik3
from kamzik3.snippets.snippetsTimer import PreciseCallbackTimer

base_log_formatter = logging.Formatter('%(asctime)s, %(name)s, %(levelname)s, %(message)s', datefmt='%Y-%m-%d %H:%M:%S')


def set_file_handler(logger, log_output_dir, formatter=base_log_formatter):
    handler = logging.FileHandler(log_output_dir)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return handler


def set_rotating_file_handler(logger, log_output_dir, formatter=base_log_formatter):
    handler = TimedRotatingFileHandler(log_output_dir, when="midnight")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return handler


def set_rocket_chat_handler(logger, main_room, discussion, user_credentials=None, auth_token=None,
                            level=logging.ERROR, formatter=base_log_formatter):
    handler = RocketChatHandler(user_credentials=user_credentials, auth_token=auth_token, main_room_name=main_room,
                                discussion_name=discussion, level=level)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return handler


def get_console_handler(log_level=logging.DEBUG, formatter=base_log_formatter):
    console_handler = logging.StreamHandler()
    console_handler.setLevel(log_level)
    console_handler.setFormatter(formatter)
    return console_handler


def set_sys_exception_handler(logger):
    """
    Pass all Uncaught exception into logger handler
    :param logger:
    :return:
    """

    def handle_exception(exc_type, exc_value, exc_traceback):
        logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))
        sys.__excepthook__(exc_type, exc_value, exc_traceback)

    sys.excepthook = handle_exception


def print_exception():
    """
    Print formatted exception in stdout
    :return: None
    """
    print(traceback.format_exc())


class RocketChatHandler(Handler):

    def __init__(self, user_credentials=None, auth_token=None, main_room_name="kamzik3", discussion_name="kamzik3",
                 server_url="https://cfel-rocketchat.desy.de", level=logging.NOTSET):

        # Import rocketchat library
        from rocketchat_API.rocketchat import RocketChat
        from requests import sessions

        self.discussion_name = discussion_name
        self.main_room_name = main_room_name

        # Check if we have any user credentials or access token
        if (user_credentials is None or (user_credentials[0] == user_credentials[1] is None)) and auth_token is None:
            raise Exception(f"No user authentication method was provided!")

        # Check if session was already open
        if not hasattr(kamzik3, "rc_session"):
            kamzik3.rc_session = sessions.Session()
            kamzik3.rc = RocketChat(*user_credentials, auth_token=auth_token, server_url=server_url,
                                    session=kamzik3.rc_session)

        # Check if Discussion already exists
        all_rooms = kamzik3.rc.rooms_get().json()
        self.main_room_id = None
        self.discussion_room_id = None
        self.available_discussions = []
        # Loop over all groups, that user has access to
        for room in all_rooms["update"]:
            # We found main room
            if room.get("name") == self.main_room_name:
                self.main_room_id = room["_id"]
                # Get all discussions within room
                room_discussions = kamzik3.rc.call_api_get("chat.getDiscussions", roomId=self.main_room_id).json()
                for discussion in room_discussions["messages"]:
                    # Discussion was found
                    if discussion["msg"] == self.discussion_name:
                        self.discussion_room_id = discussion["drid"]
                        break
                break

        # Throw Exception if main room does not exists
        if self.main_room_id is None:
            raise Exception(f"RocketChat {self.main_room_name} room does not exists or user has no access right!")
        # Throw Exception if discussion does not exists
        if self.discussion_room_id is None:
            raise Exception(
                f"RocketChat {self.discussion_name} discussion does not exists or user has no access right!")

        Handler.__init__(self, level)

    def emit(self, record):
        """
        Write logged message in RocketChat Discussion
        :param record: str
        :return: None
        """
        if record.exc_info is not None:
            msg = "```" + self.format(record) + "```" + record.getMessage()
        elif record.levelno == logging.ERROR:
            msg = "```" + traceback.format_exc() + "```" + record.getMessage()
        else:
            msg = record.getMessage()
        kamzik3.rc.chat_post_message(msg, room_id=self.discussion_room_id).json()


class MeasureTiming:

    def __init__(self, title, timeout, save_file, measure_time=10000):
        self.spacing = []
        self.error = []
        self.measured_values = []
        self.title = title
        self.measure_time = int(measure_time)
        self.samples = 0
        self.running = False
        self.timeout = timeout
        self.timer = None
        self.save_file = save_file
        self.reference_time_perf = perf_counter_ns()

    def start_measurement(self):
        self.spacing = []
        self.error = []
        self.samples = 0
        self.running = True
        self.timer = PreciseCallbackTimer(self.measure_time + 10 * self.timeout, self.plot)
        self.reference_time_perf = perf_counter_ns()
        self.timer.start()

    def measure(self, data=None):
        if not self.running:
            return
        t_diff = (perf_counter_ns() - self.reference_time_perf) * 1e-6
        self.measured_values.append(str((perf_counter_ns(), t_diff, data)))
        self.reference_time_perf = perf_counter_ns()
        self.spacing.append(t_diff)
        self.error.append(abs(t_diff - self.timeout))
        self.samples += 1

    def plot(self):
        self.running = False
        self.timer.stop()
        sleep(0.5)
        self.spacing = self.spacing[5:-5]
        self.error = self.error[5:-5]
        stat_data = []
        stat_data.append(self.title)
        stat_data.append(str((time(), "Frequency:", self.samples, "Median:", statistics.median(self.spacing),
                              "Mean:", statistics.mean(self.spacing), "Min:", min(self.spacing), "Max:",
                              max(self.spacing), "Std:",
                              numpy.std(self.spacing))))
        stat_data.append(
            str(("Median error:", statistics.median(self.error), "Mean error:", statistics.mean(self.error), "Min:",
                 min(self.error),
                 "Max:",
                 max(self.error),
                 "Std:", numpy.std(self.error))))
        print(stat_data)
        with open("./collected_data.txt", "a+") as fp:
            fp.write("\r\n".join(stat_data))
            fp.write("\r\n".join(self.measured_values))

        import matplotlib.pyplot as plt
        plt.subplot(2, 1, 1)
        plt.scatter(range(len(self.spacing)), self.spacing, s=2, c="r")
        plt.title(self.title)
        plt.xlabel('samples [count]')
        plt.ylabel('spacing [ms]')

        plt.subplot(2, 1, 2)
        plt.scatter(range(len(self.error)), self.error, s=2, c="r")
        plt.xlabel('samples [count]')
        plt.ylabel('error [ms]')

        plt.savefig(self.save_file)



import sys
from threading import Timer
from time import perf_counter_ns

from kamzik3.snippets.snippetsInterfaces import CallbackTimerInterface


class CallbackTimer(CallbackTimerInterface):

    def run(self):
        if self.with_correction:
            time_reference = perf_counter_ns()
            while not self.stopped.wait(self.timeout - (perf_counter_ns() - time_reference) * 1e-9):
                self.callback()
                time_reference += self.timeout * 1e9
        else:
            while not self.stopped.wait(self.timeout):
                self.callback()


if sys.platform == "win32":
    # If more precise timer is requested for windows.
    # Problem of windows is that it's running on lower clock compared to Linux platform.
    # We can get only to resolution of 10ms.
    # We have to increase clock of CPU.
    # THIS CHANGES CLOCK GLOBALLY !!!!!!!!
    from kamzik3.snippets.snippetsWin32 import CallbackTimerWOC


    class PreciseCallbackTimer(CallbackTimerWOC):
        pass
else:
    # For Linux we use same timer since there is no problem with clock precision.
    # We can get up to 1us resolution.
    class PreciseCallbackTimer(CallbackTimer):
        pass


class RecallableTimer:

    def __init__(self, interval, f, *args, **kwargs):
        self.interval = interval
        self.f = f
        self.args = args
        self.kwargs = kwargs

        self.timer = None

    def callback(self):
        self.f(*self.args, **self.kwargs)

    def cancel(self):
        self.timer.cancel()

    def start(self):
        self.timer = Timer(self.interval, self.callback)
        self.timer.start()



class Observer(object):
    """
    Implement this class to implement Kamzik3 Observer.
    """

    def subject_update(self, key, value, subject):
        """
        Receive an subject update.
        :param key: mixed
        :param value: mixed
        :param subject: Subject
        :return: None
        """
        raise NotImplementedError(u"Not implemented")

    def attach_to_subject(self, subject):
        """
        Connect Observer to Subject.
        :param subject: Subject
        :return: None
        """
        subject.attach_observer(self)

    def detach_subject(self, subject):
        """
        Connect Observer to Subject.
        :param subject: Subject
        :return: None
        """
        subject.detach_observer(self)
